# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""OpenSpiel support for the Atari Learning Environment (ALE).

Originally introduced in (Bellemare et al., 2013):
https://arxiv.org/abs/1207.4708.

Uses environment wrappers from OpenAI Gym (https://gym.openai.com/) and Stable
Baselines 3 (https://jmlr.org/papers/v22/20-1364.html) to convert observations
into a suitable format for training.
"""

# pylint: disable=g-importing-member
from math import prod
import gym
import numpy as np
from stable_baselines3.common.atari_wrappers import ClipRewardEnv
from stable_baselines3.common.atari_wrappers import EpisodicLifeEnv
from stable_baselines3.common.atari_wrappers import FireResetEnv
from stable_baselines3.common.atari_wrappers import MaxAndSkipEnv
import pyspiel


_NUM_PLAYERS = 1
_GAME_TYPE = pyspiel.GameType(
    short_name='atari',
    long_name='atari',
    dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
    chance_mode=pyspiel.GameType.ChanceMode.SAMPLED_STOCHASTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.ZERO_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=False,
    provides_information_state_tensor=False,
    provides_observation_string=True,
    provides_observation_tensor=True,
    parameter_specification={
        'gym_id': 'ALE/Breakout-v5',
        'seed': 1,
        'idx': 0,
        'capture_video': False,
        'run_name': 'default',
        'use_episodic_life_env': True
    })
_GAME_INFO = pyspiel.GameInfo(
    num_distinct_actions=4,
    max_chance_outcomes=0,
    num_players=_NUM_PLAYERS,
    min_utility=-1.0,
    max_utility=1.0,
    utility_sum=0.0,
    max_game_length=2000)


# NOTE: We include this wrapper by hand because the default wrapper
# threw errors (see modified lines).
class NoopResetEnv(gym.Wrapper):
  """Sample initial states by taking random number of no-ops on reset.

  No-op is assumed to be action 0. :param env: the environment to wrap :param
  noop_max: the maximum value of no-ops to run
  """

  def __init__(self, env: gym.Env, noop_max: int = 30):
    gym.Wrapper.__init__(self, env)
    self.noop_max = noop_max
    self.override_num_noops = None
    self.noop_action = 0
    assert env.unwrapped.get_action_meanings()[0] == 'NOOP'

  def reset(self, **kwargs) -> np.ndarray:
    self.env.reset(**kwargs)
    if self.override_num_noops is not None:
      noops = self.override_num_noops
    else:
      #### MODIFIED LINES: note method is named integers now ###
      noops = self.unwrapped.np_random.integers(1, self.noop_max + 1)
      ### END MODIFIED LINES ###
    assert noops > 0
    obs = np.zeros(0)
    for _ in range(noops):
      obs, _, done, _ = self.env.step(self.noop_action)
      if done:
        obs = self.env.reset(**kwargs)
    return obs


class AtariGame(pyspiel.Game):
  """An OpenSpiel wrapper for the OpenAI Gym Atari games."""

  def __init__(self, params=None):
    super().__init__(_GAME_TYPE, _GAME_INFO, params or dict())
    self.gym_id = params.get('gym_id', 'BreakoutNoFrameskip-v4')
    self.seed = params.get('seed', 1)
    self.idx = params.get('idx', 0)
    self.capture_video = params.get('capture_video', False)
    self.run_name = params.get('run_name', 'default')
    self.use_episodic_life_env = params.get('use_episodic_life_env', True)

    env = gym.make(self.gym_id)
    env = gym.wrappers.RecordEpisodeStatistics(env)
    if self.capture_video and self.idx == 0:
      env = gym.wrappers.RecordVideo(env, f'videos/{self.run_name}')

    # Apply the standard set of wrappers from CleanRL's PPO implementation.
    # These wrappers have been tested on Breakout; different games may
    # benefit from different wrappers (e.g., Space Invaders might benefit
    # from frameskip=3 instead of 4; see https://arxiv.org/abs/1312.5602).
    env = NoopResetEnv(env, noop_max=30)
    env = MaxAndSkipEnv(env, skip=4)
    if self.use_episodic_life_env:
      env = EpisodicLifeEnv(env)
    if 'FIRE' in env.unwrapped.get_action_meanings():
      env = FireResetEnv(env)
    env = ClipRewardEnv(env)
    env = gym.wrappers.ResizeObservation(env, (84, 84))
    env = gym.wrappers.GrayScaleObservation(env)
    env = gym.wrappers.FrameStack(env, 4)
    env.seed(self.seed)
    env.action_space.seed(self.seed)
    env.observation_space.seed(self.seed)
    self.observation_shape = env.reset().shape
    self.env = env

  def observation_tensor_shape(self):
    return self.observation_shape

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return AtariState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    if params is None:
      params = dict()

    params['observation_shape'] = self.observation_shape
    return AtariObserver(
        iig_obs_type or pyspiel.IIGObservationType(perfect_recall=False),
        params)


class AtariState(pyspiel.State):
  """A python version of the Atari Game state."""

  def __init__(self, game):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._is_terminal = False
    self.tracked_rewards = 0
    self.env = game.env
    self.observation = self.env.reset()
    self.last_reward = None
    self.last_info = dict()

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    return pyspiel.PlayerId.TERMINAL if self._is_terminal else 0

  def _legal_actions(self, player):
    """Returns a list of legal actions, sorted in ascending order."""
    return list(range(self.env.action_space.n))

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    observation, reward, done, info = self.env.step(action)
    self.last_info = info
    self.last_reward = reward
    self.tracked_rewards += reward
    if done:
      self._is_terminal = True
    self.observation = observation  # Store this for later

  def _action_to_string(self, player, action):
    return self.env.get_action_meanings()[action]

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._is_terminal

  def rewards(self):
    return [self.last_reward]

  def returns(self):
    """Total reward for each player over the course of the game so far."""
    return [self.tracked_rewards]

  def __str__(self):
    """String for debug purposes. No particular semantics are required."""
    return 'DEBUG'


class AtariObserver:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  # pylint: disable=unused-argument
  def __init__(self, iig_obs_type, params):
    """Initializes an empty observation tensor."""
    # Determine which observation pieces we want to include.
    pieces = []
    pieces.append(('observation', prod(params['observation_shape']),
                   params['observation_shape']))

    # Build the single flat tensor.
    total_size = sum(size for name, size, shape in pieces)
    self.tensor = np.zeros((total_size), np.float32)

    # Build the named & reshaped views of the bits of the flat tensor.
    self.dict = {}
    index = 0
    for name, size, shape in pieces:
      self.dict[name] = self.tensor[index:index + size].reshape(shape)
      index += size

  def set_from(self, state, player):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    self.tensor.fill(0)
    if 'observation' in self.dict:
      self.dict['observation'][:] = state.observation

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    pieces = []
    return ' '.join(str(p) for p in pieces)


# Register the game with the OpenSpiel library
pyspiel.register_game(_GAME_TYPE, AtariGame)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Examples of debates -- useful for generating more examples.
"""

from open_spiel.python.games.chat_games.envs.utils import text

# Scenario A
SCENARIO_A_LIST = ['Tom Brady is the GOAT and coach Bill Bellichick ' +
                   'is a genius']
SCENARIO_A = '\n\n'.join(text.wrap(SCENARIO_A_LIST))

TOPIC_A = 'The New England Patriots are the best NFL team in 2023.'

INFO_A = ''

# Scenario B
SCENARIO_B_LIST = ['Breakfast is the most important meal of the day.']
SCENARIO_B = '\n\n'.join(text.wrap(SCENARIO_B_LIST))

TOPIC_B = 'Breakfast is the most important meal of the day.'

INFO_B = ''

LLM_TERMINATION_PROMPT = None



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.voting.borda."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python.voting import base
from open_spiel.python.voting import borda


class BordaVotingTest(parameterized.TestCase):

  def test_borda_setup(self):
    method = borda.BordaVoting()
    self.assertEqual(method.name(), "borda")

  @parameterized.named_parameters(
      dict(testcase_name="uniform votes",
           votes=[["a", "b", "c"], ["a", "c", "b"], ["b", "a", "c"]],
           ranking=["a", "b", "c"],
           scores=[5, 3, 1]),
      dict(testcase_name="weighted votes",
           votes=[
               base.WeightedVote(1, ["a", "b", "c"]),
               base.WeightedVote(2, ["a", "c", "b"]),
               base.WeightedVote(3, ["b", "a", "c"])
           ],
           ranking=["a", "b", "c"],
           scores=[9, 7, 2]))
  def test_borda_basic_run(self, votes, ranking, scores):
    profile = base.PreferenceProfile(votes=votes)
    method = borda.BordaVoting()
    outcome = method.run_election(profile)
    with self.subTest("ranking correct"):
      self.assertListEqual(outcome.ranking, ranking)
    with self.subTest("scores correct"):
      self.assertListEqual(outcome.scores, scores)


if __name__ == "__main__":
  absltest.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MIP-Nash.

Based on the first formulation of
 https://dl.acm.org/doi/10.5555/1619410.1619413.
Compute optimal Nash equilibrium of two-player general-sum games
by solving a mixed-integer programming problem.
"""


import cvxpy as cp
import numpy as np
from open_spiel.python.algorithms.projected_replicator_dynamics import _simplex_projection
from open_spiel.python.egt.utils import game_payoffs_array


def mip_nash(game, objective, solver='GLPK_MI'):
  """Solves for the optimal Nash for two-player general-sum games.

    Using mixed-integer programming:
      min f(x_0, x_1, p_mat)
      s.t.
      (u_0, u_1 are Nash payoffs variables of player 0 and 1)
      p_mat[0] * x_1 <= u_0
      x_0^T*p_mat[1] <= u_1
      (if a pure strategy is in the support then its payoff is Nash payoff)
      u_0 - p_mat[0] * x_1 <= u_max_0 * b_0
      u_1 - x_0^T*p_mat[1] <= u_max_1 * b_1
      (if a pure strategy is not in the support its probability mass is 0)
      x_0 <= 1 - b_0
      x_1 <= 1 - b_1
      (probability constraints)
      x_0 >= 0
      1^T * x_0 = 1
      x_1 >= 0
      1^T * x_1 = 1
      for all n, b_0[n] in {0, 1},
      for all m, b_1[m] in {0, 1},
      u_max_0, u_max_1 are the maximum payoff differences of player 0 and 1.
    Note: this formulation is a basic one that may only work well
    for simple objective function or low-dimensional inputs.
    GLPK_MI solver only handles linear objective.
    To handle nonlinear and high-dimensional cases,
    it is recommended to use advance solvers such as GUROBI,
    or use a piecewise linear approximation of the objective.
  Args:
    game: a pyspiel matrix game object
    objective: a string representing the objective (e.g., MAX_SOCIAL_WELFARE)
    solver: the mixed-integer solver used by cvxpy

  Returns:
    optimal Nash (x_0, x_1)
  """

  p_mat = game_payoffs_array(game)
  if len(p_mat) != 2:
    raise ValueError('MIP-Nash only works for two players.')

  assert len(p_mat) == 2
  assert p_mat[0].shape == p_mat[1].shape

  (m_0, m_1) = p_mat[0].shape

  u_max_0 = np.max(p_mat[0]) - np.min(p_mat[0])
  u_max_1 = np.max(p_mat[1]) - np.min(p_mat[1])

  x_0 = cp.Variable(m_0)
  x_1 = cp.Variable(m_1)
  u_0 = cp.Variable(1)
  u_1 = cp.Variable(1)
  b_0 = cp.Variable(m_0, boolean=True)
  b_1 = cp.Variable(m_1, boolean=True)

  u_m = p_mat[0] @ x_1
  u_n = x_0 @ p_mat[1]

  # probabilities constraints
  constraints = [x_0 >= 0, x_1 >= 0, cp.sum(x_0) == 1, cp.sum(x_1) == 1]
  # support constraints
  constraints.extend([u_m <= u_0, u_0 - u_m <= u_max_0 * b_0, x_0 <= 1 - b_0])
  constraints.extend([u_n <= u_1, u_1 - u_n <= u_max_1 * b_1, x_1 <= 1 - b_1])

  variables = {
      'x_0': x_0,
      'x_1': x_1,
      'u_0': u_0,
      'u_1': u_1,
      'b_0': b_0,
      'b_1': b_1,
      'p_mat': p_mat,
  }

  obj = TWO_PLAYER_OBJECTIVE[objective](variables)
  prob = cp.Problem(obj, constraints)
  prob.solve(solver=solver)

  return _simplex_projection(x_0.value.reshape(-1)), _simplex_projection(
      x_1.value.reshape(-1)
  )


def max_social_welfare_two_player(variables):
  """Max social welfare objective."""
  return cp.Maximize(variables['u_0'] + variables['u_1'])


def min_social_welfare_two_player(variables):
  """Min social welfare objective."""
  return cp.Minimize(variables['u_0'] + variables['u_1'])


def max_support_two_player(variables):
  """Max support objective."""
  return cp.Minimize(cp.sum(variables['b_0']) + cp.sum(variables['b_1']))


def min_support_two_player(variables):
  """Min support objective."""
  return cp.Maximize(cp.sum(variables['b_0']) + cp.sum(variables['b_1']))


def max_gini_two_player(variables):
  """Max gini objective."""
  return cp.Minimize(
      cp.sum(cp.square(variables['x_0'])) + cp.sum(cp.square(variables['x_1']))
  )


TWO_PLAYER_OBJECTIVE = {
    'MAX_SOCIAL_WELFARE': max_social_welfare_two_player,
    'MIN_SOCIAL_WELFARE': min_social_welfare_two_player,
    'MAX_SUPPORT': max_support_two_player,
    'MIN_SUPPORT': min_support_two_player,
    'MAX_GINI': max_gini_two_player,
}



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mean-Field Bandit Regret Minimizers from Muller et al."""

from typing import Optional

import numpy as np
import scipy.optimize
import scipy.sparse.linalg

from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import utils


# pylint: disable=invalid-name
def get_proba_constraints_positivity(nus):
  A = np.zeros((nus.shape[0], 1 + nus.shape[0]))
  A[:, 1:] = -np.eye(nus.shape[0])
  return A, np.zeros(A.shape[0])


def get_proba_constraint_sum_eq(nus):
  A = np.ones((1, 1 + nus.shape[0]))
  A[0, 0] = 0.0
  return A, np.array([1.0])


def compress_internal_weights(nus, regrets):
  """Compress internal weights.

  Via optimization, identify which regret timesteps are useful and which aren't
  for internal regret.

  Args:
    nus: Distribution per timestep.
    regrets: Regret value per timestep and action.

  Returns:
    Weights over nus which can be used to average the no-regret distribution.
  """

  def get_c(nus):
    return np.concatenate((np.array([1.0]), np.zeros(nus.shape[0])))

  def get_max_constraint(regrets):
    regrets = np.transpose(np.array(regrets), axes=[0, 2, 1])
    regrets = regrets.reshape(-1, regrets.shape[-1])
    A = np.zeros((regrets.shape[0], 1 + regrets.shape[1]))
    A[:, 1:] = regrets
    A[:, 0] = -1.0

    b = np.zeros(A.shape[0])
    return A, b

  def get_a_ub(nus, regrets):
    Amax, bmax = get_max_constraint(regrets)
    Apos, bpos = get_proba_constraints_positivity(nus)
    return np.concatenate((Amax, Apos), axis=0), np.concatenate(
        (bmax, bpos), axis=0
    )

  c = get_c(nus)

  A_ub, b_ub = get_a_ub(nus, regrets)
  A_eq, b_eq = get_proba_constraint_sum_eq(nus)

  res = scipy.optimize.linprog(
      c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, options={"tol": 1e-10}
  )
  new_weights = res.x
  return new_weights[1:]


def compress_external_weights(nus, regrets, lbd=0.0):
  """Compress internal weights.

  Via optimization, identify which regret timesteps are useful and which aren't
  for external regret.

  Args:
    nus: Distribution per timestep.
    regrets: Regret value per timestep and action.
    lbd: Sparsity penalty.

  Returns:
    Weights over nus which can be used to average the no-regret distribution.
  """

  def get_c(nus):
    return np.concatenate((np.array([1.0]), np.zeros(nus.shape[0])))

  def get_max_constraints(nus, regrets, lbd):
    A = np.zeros((regrets.shape[1], 1 + nus.shape[0]))
    A[:, 0] = -1.0
    A[:, 1:] = np.transpose(
        regrets
        - np.sum(regrets * nus, axis=1).reshape(-1, 1)
        - lbd * np.abs(regrets)
    )
    return A, np.zeros(A.shape[0])

  def get_a_ub(nus, regrets, lbd):
    Amax, bmax = get_max_constraints(nus, regrets, lbd)
    Apos, bpos = get_proba_constraints_positivity(nus)
    return np.concatenate((Amax, Apos), axis=0), np.concatenate(
        (bmax, bpos), axis=0
    )

  c = get_c(nus)

  A_ub, b_ub = get_a_ub(nus, regrets, lbd)
  A_eq, b_eq = get_proba_constraint_sum_eq(nus)

  res = scipy.optimize.linprog(
      c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, options={"tol": 1e-10}
  )
  new_weights = res.x
  return new_weights[1:]


# Faster than using scipy.linalg.eig.
def power_method(w_nus):
  """Quick implementation of the power method.

  Args:
    w_nus:

  Returns:
    Highest eigenvalue of the system.

  Raises:
    ValueError: when the power method did not converge after 10.000 trials.
  """
  p = np.ones(len(w_nus))
  pprime = np.dot(p, w_nus)
  n_trials = 10000
  i = 0
  while np.sum(np.abs(pprime - p)) > 1e-8 and i < n_trials:
    p = pprime
    pprime = np.dot(p, w_nus)
    pprime[pprime < 0] = 0.0
    pprime /= np.sum(pprime)
    i += 1

  if np.sum(np.abs(pprime - p)) > 1e-8 and i >= n_trials:
    raise ValueError(
        "Power method did not converge after {} trials.".format(n_trials)
    )

  p[p < 0] = 0.0
  return p / np.sum(p)


class RegretMinimizer(object):
  """Base class for Regret Minimizers.

  Implements base functions for regret minimizers to implement.

  Attributes:
    _game: Pyspiel game.
    _regret_steps_per_step: Number of regret steps per `step` call (Maximum
      number in case `stop_early` is true)
    _rho_tol: If `_compress_nus` is true, minimum probability threshold (
      Probabilities below `rho_tol` will be filtered out).
    _compress_nus: Whether to compress nus (Remove nus with low selection
      probability) or not.
    _compress_lbd: Penalty term in L1 minimization when compressing nus.
    _stop_early: Whether to stop regret computation when average regret is lower
      than `_stop_regret_threshold` or to keep going until
      `_regret_steps_per_step` steps have been accomplished.
    _stop_regret_threshold: If `stop_early` is true, average regret threshold
      under which the algorithm will stop.
    _policies: List of Policies
    _value_estimator: Value estimation function.
    _value_estimation_n: Number of runs to average _value_estimator's result on.
  """

  def __init__(
      self,
      game,
      policies,
      regret_steps_per_step: int = 1,
      rho_tol: float = 1e-4,
      compress_nus: bool = True,
      compress_every: int = 1,
      compress_lbd: float = 0.0,
      stop_early: bool = True,
      stop_regret_threshold: float = 1e-3,
      value_estimator=utils.sample_value,
      value_estimation_n: int = 1,
      compute_internal_regret: bool = False,
  ):
    self._game = game
    self._regret_steps_per_step = regret_steps_per_step

    self._compress_nus = compress_nus
    self._compress_every = compress_every
    self._compress_lbd = compress_lbd

    self._stop_early = stop_early
    self._stop_regret_threshold = stop_regret_threshold

    self._rho_tol = rho_tol
    self._policies = policies

    self._value_estimator = value_estimator
    self._value_estimation_n = value_estimation_n

    self._compute_internal_regret = compute_internal_regret

  def update_policy_mus(self):
    """Update the stored distributions of our policies."""
    self._policy_mus = [
        distribution.DistributionPolicy(self._game, policy)
        for policy in self._policies
    ]

  def get_nu(self):
    """Returns current Population Distribution."""
    raise NotImplementedError

  def step(self):
    """Make a regret minimization step."""
    raise NotImplementedError

  def step_for(self, T):
    """Do `T` steps."""
    raise NotImplementedError

  def compute_average_regret(self):
    raise NotImplementedError

  def compute_regrets(self):
    raise NotImplementedError

  def reset(self, policies):
    """Restart the bandit with new policies."""
    raise NotImplementedError


def polynomial_weight_update(weights, rewards, eta):
  return weights * (1 + eta * rewards)


class PolynomialWeightAlgorithm(RegretMinimizer):
  """Implements the Polynomial Weight Algorithm Regret minimizer.

  This is an external-regret minimizer, adapted here to the Mean-Field,
  Partially-Observable case.
  """

  def __init__(
      self,
      game,
      policies,
      eta: Optional[float] = None,
      regret_steps_per_step: int = 1,
      rho_tol: float = 1e-4,
      compress_nus: bool = True,
      compress_every: int = 1,
      compress_lbd: float = 0.0,
      stop_early: bool = True,
      stop_regret_threshold: float = 1e-3,
      value_estimator=utils.sample_value,
      value_estimation_n: int = 1,
      compute_internal_regret: bool = False,
  ):
    super().__init__(
        game,
        policies,
        regret_steps_per_step=regret_steps_per_step,
        rho_tol=rho_tol,
        compress_nus=compress_nus,
        compress_every=compress_every,
        compress_lbd=compress_lbd,
        stop_early=stop_early,
        stop_regret_threshold=stop_regret_threshold,
        value_estimator=value_estimator,
        value_estimation_n=value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )
    if self._compute_internal_regret:
      self._ws = [np.ones(len(policies)) for _ in range(len(policies))]
      self._p = np.ones(len(policies)) / (1.0 * len(policies))
    else:
      self._w = np.ones(len(policies))

    if eta is None:
      assert regret_steps_per_step is not None, (
          "Both `eta` and "
          "`regret_steps_per_step` were "
          "None, whereas our algorithm "
          "requires either value to be "
          "set."
      )
      self.compute_optimal_eta()
    else:
      self._eta = eta

    self._nus = []
    self._rewards = []
    self._policy_mus = []
    self._nu_weights = []

  def get_all_w_nus(self):
    assert self._compute_internal_regret
    return [w / np.sum(w) for w in list(self._ws)]

  def get_nu(self):
    if self._compute_internal_regret:
      return np.sum(
          self._p.reshape(-1, 1) * np.array(self.get_all_w_nus()), axis=0
      )
    else:
      return self._w / np.sum(self._w)

  def compute_p(self):
    assert (
        self._compute_internal_regret
    ), "`p` does not exist when computing external regret."
    w_nus = np.array(self.get_all_w_nus())

    p = power_method(w_nus)
    self._p = p

  def _update_weights(self, rewards):
    if self._compute_internal_regret:
      self._ws = [
          w * (1 + self._eta * rewards * p) for w, p in zip(self._ws, self._p)
      ]
      self.compute_p()
    else:
      self._w = self._w * (1 + self._eta * rewards)

  def step(self):
    rewards = np.zeros(len(self._policies))
    nu = self.get_nu()
    self._nus.append(nu)
    self._nu_weights = list(self._nu_weights)
    self._nu_weights.append(1.0)

    mu = utils.MixedDistribution(self._policy_mus, nu)
    for _ in range(self._value_estimation_n):
      for index, policy in enumerate(self._policies):
        rewards[index] += self._value_estimator(policy, mu, self._game)
    rewards /= self._value_estimation_n

    self._update_weights(rewards)
    self._rewards.append(rewards)

  def step_for(self, T):
    if self._compute_internal_regret:
      print("Minimizing Internal Regret")
    else:
      print("Minimizing External Regret")
    for t in range(T):
      self.step()
      if self._stop_early and (t % self._compress_every == 0):
        try:
          regret, weights = self.get_post_compression_regret_and_weights()
          # print("{}".format(regret))
          assert np.abs(np.sum(weights) - 1.0) < 1e-8
        except:  # pylint: disable=bare-except
          print("Simplex method encountered an error.")
          continue
        if regret < self._stop_regret_threshold:
          break
    self.compress_nus_and_weights(weights)

  def get_post_compression_regret_and_weights(self):
    """Compress the regret and weights."""
    if self._compute_internal_regret:
      nu_weights = compress_internal_weights(
          self.get_nus(), self.compute_regrets()
      )
      regret = np.max([
          np.max(np.sum(nu_weights.reshape(-1, 1) * a, axis=0))
          for a in self.compute_regrets()
      ])
    else:
      nu_weights = compress_external_weights(
          self.get_nus(), self.compute_regrets(), lbd=self._compress_lbd
      )
      regret = np.max(
          np.sum(nu_weights.reshape(-1, 1) * self.compute_regrets(), axis=0)
      )
    return regret, nu_weights

  def compress_nus_and_weights(self, nu_weights):
    """Run L1 optimization to only keep important members of `nus`."""
    if self._compress_nus:
      try:
        assert np.abs(np.sum(nu_weights) - 1.0) < 1e-8
      except:  # pylint: disable=bare-except
        # If the optimization was unsuccessful, do *not* compress.
        return

      new_nus = [
          nu
          for weight, nu in zip(nu_weights, self._nus)
          if weight > self._rho_tol
      ]
      new_rewards = [
          reward
          for weight, reward in zip(nu_weights, self._rewards)
          if weight > self._rho_tol
      ]
      new_nu_weights = [
          weight for weight in nu_weights if weight > self._rho_tol
      ]
      new_nu_weights = np.array(new_nu_weights) / np.sum(new_nu_weights)

      self._nus = new_nus
      self._rewards = new_rewards
      self._nu_weights = new_nu_weights

  def normalize_nu_weights(self):
    self._nu_weights = np.array(self._nu_weights) / np.sum(self._nu_weights)

  def get_normalized_nu_weights(self):
    return np.array(self._nu_weights) / np.sum(self._nu_weights)

  def compute_regrets(self):
    if self._compute_internal_regret:
      regrets = []
      nus = np.array(self._nus)
      rewards = np.array(self._rewards)
      for action in range(rewards.shape[1]):
        on_policy_values = (rewards[:, action] * nus[:, action]).reshape(-1, 1)
        action_values = rewards * nus[:, action].reshape(-1, 1)
        regrets.append(action_values - on_policy_values)
    else:
      on_policy_value = np.sum(
          self._rewards * np.array(self._nus), axis=1, keepdims=True
      )
      policy_value = self._rewards
      regrets = policy_value - on_policy_value
    return regrets

  def compute_average_regret(self):
    nu_weights = self.get_normalized_nu_weights()
    if self._compute_internal_regret:
      regrets = 0.0
      nus = np.array(self._nus)
      rewards = np.array(self._rewards)
      for action in range(rewards.shape[1]):
        on_policy_values = (rewards[:, action] * nus[:, action]).reshape(-1, 1)
        action_values = rewards * nus[:, action].reshape(-1, 1)
        regrets += np.max(
            np.sum(
                nu_weights.reshape(-1, 1) * (action_values - on_policy_values),
                axis=0,
            )
        )
    else:
      regrets = np.sum(
          nu_weights.reshape(-1, 1) * self.compute_regrets(), axis=0
      )
    return np.max(regrets) / len(self._nus)

  def get_nus(self):
    return np.array(self._nus)

  def get_mus(self):
    mus = []
    for nu in self._nus:
      mu = utils.MixedDistribution(self._policy_mus, nu)
      mus.append(mu)
    return mus

  def get_rewards(self):
    return self._rewards

  def get_mus_and_weights(self):
    mus = self.get_mus()
    self.normalize_nu_weights()
    return mus, self._nu_weights

  def compute_optimal_eta(self):
    if self._regret_steps_per_step is not None:
      self._eta = min(
          np.sqrt(np.log(len(self._policies)) / self._regret_steps_per_step),
          0.5,
      )

  def reset(self, policies):
    if self._compute_internal_regret:
      self._ws = [np.ones(len(policies)) for _ in range(len(policies))]
      self._p = np.ones(len(policies)) / (1.0 * len(policies))
    else:
      self._w = np.ones(len(policies))
    self._policies = policies
    self._nus = []
    self._rewards = []
    self._policy_mus = []
    self._nu_weights = []
    self.update_policy_mus()
    self.compute_optimal_eta()


class Hedge(PolynomialWeightAlgorithm):
  """Hedge algorithm implementation."""

  def __init__(
      self,
      game,
      policies,
      eta: Optional[float] = None,
      regret_steps_per_step: int = 1,
      rho_tol: float = 1e-4,
      compress_nus: bool = True,
      compress_lbd: float = 0.0,
      compress_every: int = 1,
      stop_early: bool = True,
      stop_regret_threshold: float = 1e-3,
      value_estimator=utils.sample_value,
      value_estimation_n: int = 1,
      compute_internal_regret: bool = False,
  ):
    super().__init__(
        game,
        policies,
        eta=eta,
        regret_steps_per_step=regret_steps_per_step,
        rho_tol=rho_tol,
        compress_nus=compress_nus,
        compress_lbd=compress_lbd,
        stop_early=stop_early,
        stop_regret_threshold=stop_regret_threshold,
        value_estimator=value_estimator,
        value_estimation_n=value_estimation_n,
        compute_internal_regret=compute_internal_regret,
    )

    if self._compute_internal_regret:
      self._ws = [np.ones(len(policies)) for _ in range(len(policies))]
      self._p = np.ones(len(policies)) / (1.0 * len(policies))
    else:
      self._w = np.ones(len(policies))

    if eta is None:
      assert regret_steps_per_step is not None, (
          "Both `eta` and "
          "`regret_steps_per_step` were "
          "None, whereas our algorithm "
          "requires either value to be "
          "set."
      )
      self.compute_optimal_eta()
    else:
      self._eta = eta

    self._compress_every = compress_every

    self._nus = []
    self._rewards = []
    self._policy_mus = []
    self._nu_weights = []

  def _update_weights(self, rewards):
    if self._compute_internal_regret:
      self._ws = [
          w * np.exp(self._eta * rewards * p) for w, p in zip(self._ws, self._p)
      ]
      self.compute_p()
    else:
      self._w = self._w * np.exp(self._eta * rewards)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Mean Field Normal Form Games / Static Mean-Field Games."""

from typing import Any, List, Mapping

import numpy as np

from open_spiel.python.observation import IIGObserverForPublicInfoGame  # pylint:disable=g-importing-member
import pyspiel


def coop_reward(last_action, distribution):
  """A game incentivising cooperation."""
  nu_a, nu_b, nu_c, *_ = distribution
  if last_action == 0:
    return 10 * nu_a - 200 / 9 * (nu_a - nu_c) * nu_c - 20 * nu_b
  elif last_action == 1:
    return 20 * (nu_a - nu_b) - 2380 * nu_c
  elif last_action == 2:
    return 2000 / 9 * (nu_a - nu_c) * nu_c
  else:
    raise ValueError("Unknown last action " + str(last_action))


def biased_indirect_rps(last_action, distribution):
  """Biased indirect Rock Paper Scissors."""
  nu_a = 0.7 * distribution[0]
  nu_b = 0.5 * distribution[1]
  nu_c = 0.3 * distribution[2]
  if last_action == 0:
    return nu_b - nu_c
  elif last_action == 1:
    return nu_c - nu_a
  elif last_action == 2:
    return nu_a - nu_b
  else:
    raise ValueError("Unknown last action " + str(last_action))


def dominated_reward_source(last_action, distribution):
  nu_a, nu_b, nu_c, *_ = distribution
  if last_action == 0:
    return nu_a + nu_c
  elif last_action == 1:
    return nu_b
  elif last_action == 2:
    return nu_a + nu_c - 0.25
  else:
    raise ValueError("Unknown last action " + str(last_action))


_NUM_PLAYERS = 1
_NUM_ACTIONS = 3
_DEFAULT_PARAMS = {"num_actions": _NUM_ACTIONS, "reward_function": "coop"}
_GAME_TYPE = pyspiel.GameType(
    short_name="mean_field_nfg",
    long_name="Mean-Field Normal-Form Game",
    dynamics=pyspiel.GameType.Dynamics.MEAN_FIELD,
    chance_mode=pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC,
    information=pyspiel.GameType.Information.PERFECT_INFORMATION,
    utility=pyspiel.GameType.Utility.GENERAL_SUM,
    reward_model=pyspiel.GameType.RewardModel.REWARDS,
    max_num_players=_NUM_PLAYERS,
    min_num_players=_NUM_PLAYERS,
    provides_information_state_string=True,
    provides_information_state_tensor=False,
    provides_observation_string=True,
    provides_observation_tensor=False,
    parameter_specification=_DEFAULT_PARAMS,
)


class MFGNormalFormGame(pyspiel.Game):
  """A Mean Field Normal Form game.

  A game starts by an initial chance node that select the initial state
  of the MFG.
  Then the game sequentially alternates between:
    - An action selection node (Where the player Id >= 0)
    - A chance node (the player id is pyspiel.PlayerId.CHANCE)
    - A Mean Field node (the player id is pyspiel.PlayerId.MEAN_FIELD)
  """

  # pylint:disable=dangerous-default-value
  def __init__(self, params: Mapping[str, Any] = _DEFAULT_PARAMS):
    game_info = pyspiel.GameInfo(
        num_distinct_actions=_NUM_ACTIONS,
        max_chance_outcomes=_NUM_ACTIONS,
        num_players=_NUM_PLAYERS,
        min_utility=-np.inf,
        max_utility=+np.inf,
        utility_sum=0.0,
        max_game_length=2,
    )
    super().__init__(_GAME_TYPE, game_info, params)
    if params["reward_function"] == "coop":
      self.reward_function = coop_reward
    elif params["reward_function"] == "dom":
      self.reward_function = dominated_reward_source
    elif params["reward_function"] == "biased_indirect_rps":
      self.reward_function = biased_indirect_rps
    else:
      raise ValueError("Unknown reward function " + params["reward_function"])
    self.num_actions = params["num_actions"]
    self.size = 1 + self.num_actions

  def new_initial_state(self):
    """Returns a state corresponding to the start of a game."""
    return MFGNormalFormState(self)

  def make_py_observer(self, iig_obs_type=None, params=None):
    """Returns an object used for observing game state."""
    if (iig_obs_type is None) or (
        iig_obs_type.public_info and not iig_obs_type.perfect_recall
    ):
      return Observer(params, self)
    return IIGObserverForPublicInfoGame(iig_obs_type, params)

  def max_chance_nodes_in_history(self):
    """Maximun chance nodes in game history."""
    return 0


class MFGNormalFormState(pyspiel.State):
  """A Mean Field Normal-Form state."""

  def __init__(self, game, last_action=None):
    """Constructor; should only be called by Game.new_initial_state."""
    super().__init__(game)
    self._player_id = pyspiel.PlayerId.DEFAULT_PLAYER_ID
    self._last_action = last_action
    self._num_actions = game.num_actions
    self.reward_function = game.reward_function
    self.size = game.size
    self._terminal = False

    # Represents the current probability distribution over game states.
    # Initialized with a uniform distribution.
    self._distribution = [1.0 / self.size for _ in range(self.size)]

  def state_to_str(self, player_id=pyspiel.PlayerId.DEFAULT_PLAYER_ID):
    """A string that uniquely identify a triplet x, t, player_id."""
    if self._last_action is None:
      return "initial"
    else:
      bonus = "_final" if self.is_terminal() else ""
      return str(self._last_action) + bonus

  # OpenSpiel (PySpiel) API functions are below. This is the standard set that
  # should be implemented by every perfect-information sequential-move game.

  def _legal_actions(self, player):
    """Returns a list of legal actions for player and MFG nodes."""
    if player == pyspiel.PlayerId.MEAN_FIELD:
      return []
    if (
        player == pyspiel.PlayerId.DEFAULT_PLAYER_ID
        and player == self.current_player()
    ):
      return list(range(self._num_actions))
    raise ValueError(
        f"Unexpected player {player}. "
        "Expected a mean field or current player 0."
    )

  def _apply_action(self, action):
    """Applies the specified action to the state."""
    if self._player_id == pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "_apply_action should not be called at a MEAN_FIELD state."
      )
    self.return_value = self._rewards()

    assert self._player_id == 0
    # Here the action is between 0 and N-1
    if action < 0 or action > self._num_actions - 1:
      raise ValueError(
          "The action is between 0 and {} at any node".format(
              self._num_actions - 1
          )
      )
    self._last_action = action
    self._player_id = pyspiel.PlayerId.MEAN_FIELD

  def _action_to_string(self, player, action):
    """Action -> string."""
    del player
    return str(action)

  def distribution_support(self):
    """return a list of state string."""
    if self._player_id == pyspiel.PlayerId.DEFAULT_PLAYER_ID:
      return [self.state_to_str()]
    elif self._player_id == pyspiel.PlayerId.MEAN_FIELD:
      return [str(i) for i in range(self._num_actions)]

  def update_distribution(self, distribution):
    """This function is central and specific to the logic of the MFG.

    Args:
      distribution: a distribution to register.  - function should be called
        when the node is in MEAN_FIELD state. - distribution are probabilities
        that correspond to each game state given by distribution_support.
    """
    if self._player_id != pyspiel.PlayerId.MEAN_FIELD:
      raise ValueError(
          "update_distribution should only be called at a MEAN_FIELD state."
      )
    self._distribution = distribution.copy()
    self._player_id = pyspiel.PlayerId.TERMINAL

  def is_terminal(self):
    """Returns True if the game is over."""
    return self._player_id == pyspiel.PlayerId.TERMINAL

  def current_player(self):
    """Returns id of the next player to move, or TERMINAL if game is over."""
    if self.is_terminal():
      return pyspiel.PlayerId.TERMINAL
    return self._player_id

  def _rewards(self):
    """Reward for the player for this state."""
    reward = 0.0
    if self._player_id == pyspiel.PlayerId.TERMINAL:
      reward = self.reward_function(self._last_action, self._distribution)
    return reward

  def rewards(self) -> List[float]:
    """Rewards for all players."""
    # For now, only single-population (single-player) mean field games
    # are supported.
    return [self._rewards()]

  def _returns(self):
    """Returns is the sum of all payoffs collected so far."""
    return self._rewards()

  def returns(self) -> List[float]:
    """Returns for all players."""
    # For now, only single-population (single-player) mean field games
    # are supported.
    return [self._returns()]

  def __str__(self):
    """A string that uniquely identify the current state."""
    return self.state_to_str(player_id=self._player_id)


class Observer:
  """Observer, conforming to the PyObserver interface (see observation.py)."""

  def __init__(self, params, game):
    """Initializes an empty observation tensor."""
    del params

    self.size = game.size
    # +1 to allow t == horizon.
    self.tensor = np.array([])
    self.dict = {}

  def set_from(self, state: MFGNormalFormState, player: int):
    """Updates `tensor` and `dict` to reflect `state` from PoV of `player`."""
    del player
    # We update the observation via the shaped tensor since indexing is more
    # convenient than with the 1-D tensor. Both are views onto the same memory.
    del state
    self.tensor.fill(0)
    # state.x is None for the initial (blank) state, don't set any
    # position bit in that case.
    pass

  def string_from(self, state, player):
    """Observation of `state` from the PoV of `player`, as a string."""
    del player
    return str(state)


# Register the game with the OpenSpiel library

pyspiel.register_game(_GAME_TYPE, MFGNormalFormGame)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Contributed by Wannes Meert, Giuseppe Marra, and Pieter Robberechts
# for the KU Leuven course Machine Learning: Project.


"""Python spiel example."""

from absl import app
from absl import flags
import numpy as np

from open_spiel.python.bots import human
from open_spiel.python.bots import uniform_random
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_integer("seed", 12761381, "The seed to use for the RNG.")

# Supported types of players: "random", "human"
flags.DEFINE_string("player0", "random", "Type of the agent for player 0.")
flags.DEFINE_string("player1", "random", "Type of the agent for player 1.")


def LoadAgent(agent_type, player_id, rng):
  """Return a bot based on the agent type."""
  if agent_type == "random":
    return uniform_random.UniformRandomBot(player_id, rng)
  elif agent_type == "human":
    return human.HumanBot()
  else:
    raise RuntimeError("Unrecognized agent type: {}".format(agent_type))


def main(_):
  rng = np.random.RandomState(FLAGS.seed)
  games_list = pyspiel.registered_names()
  assert "dots_and_boxes" in games_list

  game_string = "dots_and_boxes(num_rows=2,num_cols=2)"
  print("Creating game: {}".format(game_string))
  game = pyspiel.load_game(game_string)

  agents = [
      LoadAgent(FLAGS.player0, 0, rng),
      LoadAgent(FLAGS.player1, 1, rng),
  ]

  state = game.new_initial_state()

  # Print the initial state
  print("INITIAL STATE")
  print(str(state))

  while not state.is_terminal():
    current_player = state.current_player()
    # Decision node: sample action for the single current player
    legal_actions = state.legal_actions()
    for action in legal_actions:
      print(
          "Legal action: {} ({})".format(
              state.action_to_string(current_player, action), action
          )
      )
    action = agents[current_player].step(state)
    action_string = state.action_to_string(current_player, action)
    print("Player ", current_player, ", chose action: ", action_string)
    state.apply_action(action)

    print("")
    print("NEXT STATE:")
    print(str(state))
    if not state.is_terminal():
      print(str(state.observation_tensor()))

  # Game is now done. Print utilities for each player
  returns = state.returns()
  for pid in range(game.num_players()):
    print("Utility for player {} is {}".format(pid, returns[pid]))


if __name__ == "__main__":
  app.run(main)



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for Shapley value calculations."""


from absl.testing import absltest
import numpy as np
from open_spiel.python.coalitional_games import basic_games
from open_spiel.python.coalitional_games import deon_larson20_games
from open_spiel.python.coalitional_games import shapley_values


SEED = 23856711


class ShapleyValuesTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    np.random.seed(SEED)

  def test_ice_cream_game(self):
    """Example 2.11 from CACGT book by Chalkiadakis, Elkind, and Wooldridge."""
    game = basic_games.IceCreamGame()
    svals = shapley_values.compute_shapley_values(game)
    self.assertAlmostEqual(svals[0], 250.0)

  def test_ice_cream_game_approximate(self):
    """Monte Carlo sampling version of Shapley value computation."""
    game = basic_games.IceCreamGame()
    svals = shapley_values.compute_approximate_shapley_values(game, 1000)
    self.assertAlmostEqual(svals[0]/1000.0, 0.250, places=2)

  def test_deon_larson20_games(self):
    for name, values in deon_larson20_games.SHAPLEY_VALUES.items():
      values_arr = np.asarray(values)
      game = deon_larson20_games.make_game(name)
      svals = shapley_values.compute_shapley_values(game)
      self.assertTrue(np.allclose(svals, values_arr))


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The Dopamine Authors and Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
# This file is a fork of the original Dopamine code incorporating changes for
# the multiplayer setting and the Hanabi Learning Environment.
#
"""Implementation of a Rainbow agent adapted to the multiplayer setting."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools

import dqn_agent
import gin.tf
import numpy as np
import prioritized_replay_memory
import tensorflow as tf


slim = tf.contrib.slim


@gin.configurable
def rainbow_template(state,
                     num_actions,
                     num_atoms=51,
                     layer_size=512,
                     num_layers=1):
  r"""Builds a Rainbow Network mapping states to value distributions.

  Args:
    state: A `tf.placeholder` for the RL state.
    num_actions: int, number of actions that the RL agent can take.
    num_atoms: int, number of atoms to approximate the distribution with.
    layer_size: int, number of hidden units per layer.
    num_layers: int, number of hidden layers.

  Returns:
    net: A `tf.Graphdef` for Rainbow:
      `\theta : \mathcal{X}\rightarrow\mathbb{R}^{|\mathcal{A}| \times N}`,
      where `N` is num_atoms.
  """
  weights_initializer = slim.variance_scaling_initializer(
      factor=1.0 / np.sqrt(3.0), mode='FAN_IN', uniform=True)

  net = tf.cast(state, tf.float32)
  net = tf.squeeze(net, axis=2)

  for _ in range(num_layers):
    net = slim.fully_connected(net, layer_size,
                               activation_fn=tf.nn.relu)
  net = slim.fully_connected(net, num_actions * num_atoms, activation_fn=None,
                             weights_initializer=weights_initializer)
  net = tf.reshape(net, [-1, num_actions, num_atoms])
  return net


@gin.configurable
class RainbowAgent(dqn_agent.DQNAgent):
  """A compact implementation of the multiplayer Rainbow agent."""

  @gin.configurable
  def __init__(self,
               num_actions=None,
               observation_size=None,
               num_players=None,
               num_atoms=51,
               vmax=25.,
               gamma=0.99,
               update_horizon=1,
               min_replay_history=500,
               update_period=4,
               target_update_period=500,
               epsilon_train=0.0,
               epsilon_eval=0.0,
               epsilon_decay_period=1000,
               learning_rate=0.000025,
               optimizer_epsilon=0.00003125,
               tf_device='/cpu:*'):
    """Initializes the agent and constructs its graph.

    Args:
      num_actions: int, number of actions the agent can take at any state.
      observation_size: int, size of observation vector.
      num_players: int, number of players playing this game.
      num_atoms: Int, the number of buckets for the value function distribution.
      vmax: float, maximum return predicted by a value distribution.
      gamma: float, discount factor as commonly used in the RL literature.
      update_horizon: int, horizon at which updates are performed, the 'n' in
        n-step update.
      min_replay_history: int, number of stored transitions before training.
      update_period: int, period between DQN updates.
      target_update_period: int, update period for the target network.
      epsilon_train: float, final epsilon for training.
      epsilon_eval: float, epsilon during evaluation.
      epsilon_decay_period: int, number of steps for epsilon to decay.
      learning_rate: float, learning rate for the optimizer.
      optimizer_epsilon: float, epsilon for Adam optimizer.
      tf_device: str, Tensorflow device on which to run computations.
    """
    # We need this because some tools convert round floats into ints.
    vmax = float(vmax)
    self.num_atoms = num_atoms
    # Using -vmax as the minimum return is is wasteful, because all rewards are
    # positive -- but does not unduly affect performance.
    self.support = tf.linspace(-vmax, vmax, num_atoms)
    self.learning_rate = learning_rate
    self.optimizer_epsilon = optimizer_epsilon

    graph_template = functools.partial(rainbow_template, num_atoms=num_atoms)
    super(RainbowAgent, self).__init__(
        num_actions=num_actions,
        observation_size=observation_size,
        num_players=num_players,
        gamma=gamma,
        update_horizon=update_horizon,
        min_replay_history=min_replay_history,
        update_period=update_period,
        target_update_period=target_update_period,
        epsilon_train=epsilon_train,
        epsilon_eval=epsilon_eval,
        epsilon_decay_period=epsilon_decay_period,
        graph_template=graph_template,
        tf_device=tf_device)
    tf.logging.info('\t learning_rate: %f', learning_rate)
    tf.logging.info('\t optimizer_epsilon: %f', optimizer_epsilon)

  def _build_replay_memory(self, use_staging):
    """Creates the replay memory used by the agent.

    Rainbow uses prioritized replay.

    Args:
      use_staging: bool, whether to use a staging area in the replay memory.

    Returns:
      A replay memory object.
    """
    return prioritized_replay_memory.WrappedPrioritizedReplayMemory(
        num_actions=self.num_actions,
        observation_size=self.observation_size,
        stack_size=1,
        use_staging=use_staging,
        update_horizon=self.update_horizon,
        gamma=self.gamma)

  def _reshape_networks(self):
    # self._q is actually logits now, rename things.
    # size of _logits: 1 x num_actions x num_atoms
    self._logits = self._q
    # size of _probabilities: 1 x num_actions x num_atoms
    self._probabilities = tf.contrib.layers.softmax(self._q)
    # size of _q: 1 x num_actions
    self._q = tf.reduce_sum(self.support * self._probabilities, axis=2)
    # Recompute argmax from q values. Ignore illegal actions.
    self._q_argmax = tf.argmax(self._q + self.legal_actions_ph, axis=1)[0]

    # size of _replay_logits: 1 x num_actions x num_atoms
    self._replay_logits = self._replay_qs
    # size of _replay_next_logits: 1 x num_actions x num_atoms
    self._replay_next_logits = self._replay_next_qt
    del self._replay_qs
    del self._replay_next_qt

  def _build_target_distribution(self):
    self._reshape_networks()
    batch_size = tf.shape(self._replay.rewards)[0]
    # size of rewards: batch_size x 1
    rewards = self._replay.rewards[:, None]
    # size of tiled_support: batch_size x num_atoms
    tiled_support = tf.tile(self.support, [batch_size])
    tiled_support = tf.reshape(tiled_support, [batch_size, self.num_atoms])
    # size of target_support: batch_size x num_atoms

    is_terminal_multiplier = 1. - tf.cast(self._replay.terminals, tf.float32)
    # Incorporate terminal state to discount factor.
    # size of gamma_with_terminal: batch_size x 1
    gamma_with_terminal = self.cumulative_gamma * is_terminal_multiplier
    gamma_with_terminal = gamma_with_terminal[:, None]

    target_support = rewards + gamma_with_terminal * tiled_support
    # size of next_probabilities: batch_size  x num_actions x num_atoms
    next_probabilities = tf.contrib.layers.softmax(
        self._replay_next_logits)

    # size of next_qt: 1 x num_actions
    next_qt = tf.reduce_sum(self.support * next_probabilities, 2)
    # size of next_qt_argmax: 1 x batch_size
    next_qt_argmax = tf.argmax(
        next_qt + self._replay.next_legal_actions, axis=1)[:, None]
    batch_indices = tf.range(tf.to_int64(batch_size))[:, None]
    # size of next_qt_argmax: batch_size x 2
    next_qt_argmax = tf.concat([batch_indices, next_qt_argmax], axis=1)
    # size of next_probabilities: batch_size x num_atoms
    next_probabilities = tf.gather_nd(next_probabilities, next_qt_argmax)
    return project_distribution(target_support, next_probabilities,
                                self.support)

  def _build_train_op(self):
    """Builds the training op for Rainbow.

    Returns:
      train_op: An op performing one step of training.
    """
    target_distribution = tf.stop_gradient(self._build_target_distribution())

    # size of indices: batch_size x 1.
    indices = tf.range(tf.shape(self._replay_logits)[0])[:, None]
    # size of reshaped_actions: batch_size x 2.
    reshaped_actions = tf.concat([indices, self._replay.actions[:, None]], 1)
    # For each element of the batch, fetch the logits for its selected action.
    chosen_action_logits = tf.gather_nd(self._replay_logits, reshaped_actions)

    loss = tf.nn.softmax_cross_entropy_with_logits(
        labels=target_distribution,
        logits=chosen_action_logits)

    optimizer = tf.train.AdamOptimizer(
        learning_rate=self.learning_rate,
        epsilon=self.optimizer_epsilon)

    update_priorities_op = self._replay.tf_set_priority(
        self._replay.indices, tf.sqrt(loss + 1e-10))

    target_priorities = self._replay.tf_get_priority(self._replay.indices)
    target_priorities = tf.math.add(target_priorities, 1e-10)
    target_priorities = 1.0 / tf.sqrt(target_priorities)
    target_priorities /= tf.reduce_max(target_priorities)

    weighted_loss = target_priorities * loss

    with tf.control_dependencies([update_priorities_op]):
      return optimizer.minimize(tf.reduce_mean(weighted_loss)), weighted_loss


def project_distribution(supports, weights, target_support,
                         validate_args=False):
  """Projects a batch of (support, weights) onto target_support.

  Based on equation (7) in (Bellemare et al., 2017):
    https://arxiv.org/abs/1707.06887
  In the rest of the comments we will refer to this equation simply as Eq7.

  This code is not easy to digest, so we will use a running example to clarify
  what is going on, with the following sample inputs:
    * supports =       [[0, 2, 4, 6, 8],
                        [1, 3, 4, 5, 6]]
    * weights =        [[0.1, 0.6, 0.1, 0.1, 0.1],
                        [0.1, 0.2, 0.5, 0.1, 0.1]]
    * target_support = [4, 5, 6, 7, 8]
  In the code below, comments preceded with 'Ex:' will be referencing the above
  values.

  Args:
    supports: Tensor of shape (batch_size, num_dims) defining supports for the
      distribution.
    weights: Tensor of shape (batch_size, num_dims) defining weights on the
      original support points. Although for the CategoricalDQN agent these
      weights are probabilities, it is not required that they are.
    target_support: Tensor of shape (num_dims) defining support of the projected
      distribution. The values must be monotonically increasing. Vmin and Vmax
      will be inferred from the first and last elements of this tensor,
      respectively. The values in this tensor must be equally spaced.
    validate_args: Whether we will verify the contents of the
      target_support parameter.

  Returns:
    A Tensor of shape (batch_size, num_dims) with the projection of a batch of
    (support, weights) onto target_support.

  Raises:
    ValueError: If target_support has no dimensions, or if shapes of supports,
      weights, and target_support are incompatible.
  """
  target_support_deltas = target_support[1:] - target_support[:-1]
  # delta_z = `\Delta z` in Eq7.
  delta_z = target_support_deltas[0]
  validate_deps = []
  supports.shape.assert_is_compatible_with(weights.shape)
  supports[0].shape.assert_is_compatible_with(target_support.shape)
  target_support.shape.assert_has_rank(1)
  if validate_args:
    # Assert that supports and weights have the same shapes.
    validate_deps.append(
        tf.Assert(
            tf.reduce_all(tf.equal(tf.shape(supports), tf.shape(weights))),
            [supports, weights]))
    # Assert that elements of supports and target_support have the same shape.
    validate_deps.append(
        tf.Assert(
            tf.reduce_all(
                tf.equal(tf.shape(supports)[1], tf.shape(target_support))),
            [supports, target_support]))
    # Assert that target_support has a single dimension.
    validate_deps.append(
        tf.Assert(
            tf.equal(tf.size(tf.shape(target_support)), 1), [target_support]))
    # Assert that the target_support is monotonically increasing.
    validate_deps.append(
        tf.Assert(tf.reduce_all(target_support_deltas > 0), [target_support]))
    # Assert that the values in target_support are equally spaced.
    validate_deps.append(
        tf.Assert(
            tf.reduce_all(tf.equal(target_support_deltas, delta_z)),
            [target_support]))

  with tf.control_dependencies(validate_deps):
    # Ex: `v_min, v_max = 4, 8`.
    v_min, v_max = target_support[0], target_support[-1]
    # Ex: `batch_size = 2`.
    batch_size = tf.shape(supports)[0]
    # `N` in Eq7.
    # Ex: `num_dims = 5`.
    num_dims = tf.shape(target_support)[0]
    # clipped_support = `[\hat{T}_{z_j}]^{V_max}_{V_min}` in Eq7.
    # Ex: `clipped_support = [[[ 4.  4.  4.  6.  8.]]
    #                         [[ 4.  4.  4.  5.  6.]]]`.
    clipped_support = tf.clip_by_value(supports, v_min, v_max)[:, None, :]
    # Ex: `tiled_support = [[[[ 4.  4.  4.  6.  8.]
    #                         [ 4.  4.  4.  6.  8.]
    #                         [ 4.  4.  4.  6.  8.]
    #                         [ 4.  4.  4.  6.  8.]
    #                         [ 4.  4.  4.  6.  8.]]
    #                        [[ 4.  4.  4.  5.  6.]
    #                         [ 4.  4.  4.  5.  6.]
    #                         [ 4.  4.  4.  5.  6.]
    #                         [ 4.  4.  4.  5.  6.]
    #                         [ 4.  4.  4.  5.  6.]]]]`.
    tiled_support = tf.tile([clipped_support], [1, 1, num_dims, 1])
    # Ex: `reshaped_target_support = [[[ 4.]
    #                                  [ 5.]
    #                                  [ 6.]
    #                                  [ 7.]
    #                                  [ 8.]]
    #                                 [[ 4.]
    #                                  [ 5.]
    #                                  [ 6.]
    #                                  [ 7.]
    #                                  [ 8.]]]`.
    reshaped_target_support = tf.tile(target_support[:, None], [batch_size, 1])
    reshaped_target_support = tf.reshape(reshaped_target_support,
                                         [batch_size, num_dims, 1])
    # numerator = `|clipped_support - z_i|` in Eq7.
    # Ex: `numerator = [[[[ 0.  0.  0.  2.  4.]
    #                     [ 1.  1.  1.  1.  3.]
    #                     [ 2.  2.  2.  0.  2.]
    #                     [ 3.  3.  3.  1.  1.]
    #                     [ 4.  4.  4.  2.  0.]]
    #                    [[ 0.  0.  0.  1.  2.]
    #                     [ 1.  1.  1.  0.  1.]
    #                     [ 2.  2.  2.  1.  0.]
    #                     [ 3.  3.  3.  2.  1.]
    #                     [ 4.  4.  4.  3.  2.]]]]`.
    numerator = tf.abs(tiled_support - reshaped_target_support)
    quotient = 1 - (numerator / delta_z)
    # clipped_quotient = `[1 - numerator / (\Delta z)]_0^1` in Eq7.
    # Ex: `clipped_quotient = [[[[ 1.  1.  1.  0.  0.]
    #                            [ 0.  0.  0.  0.  0.]
    #                            [ 0.  0.  0.  1.  0.]
    #                            [ 0.  0.  0.  0.  0.]
    #                            [ 0.  0.  0.  0.  1.]]
    #                           [[ 1.  1.  1.  0.  0.]
    #                            [ 0.  0.  0.  1.  0.]
    #                            [ 0.  0.  0.  0.  1.]
    #                            [ 0.  0.  0.  0.  0.]
    #                            [ 0.  0.  0.  0.  0.]]]]`.
    clipped_quotient = tf.clip_by_value(quotient, 0, 1)
    # Ex: `weights = [[ 0.1  0.6  0.1  0.1  0.1]
    #                 [ 0.1  0.2  0.5  0.1  0.1]]`.
    weights = weights[:, None, :]
    # inner_prod = `\sum_{j=0}^{N-1} clipped_quotient * p_j(x', \pi(x'))`
    # in Eq7.
    # Ex: `inner_prod = [[[[ 0.1  0.6  0.1  0.  0. ]
    #                      [ 0.   0.   0.   0.  0. ]
    #                      [ 0.   0.   0.   0.1 0. ]
    #                      [ 0.   0.   0.   0.  0. ]
    #                      [ 0.   0.   0.   0.  0.1]]
    #                     [[ 0.1  0.2  0.5  0.  0. ]
    #                      [ 0.   0.   0.   0.1 0. ]
    #                      [ 0.   0.   0.   0.  0.1]
    #                      [ 0.   0.   0.   0.  0. ]
    #                      [ 0.   0.   0.   0.  0. ]]]]`.
    inner_prod = clipped_quotient * weights
    # Ex: `projection = [[ 0.8 0.0 0.1 0.0 0.1]
    #                    [ 0.8 0.1 0.1 0.0 0.0]]`.
    projection = tf.reduce_sum(inner_prod, 3)
    projection = tf.reshape(projection, [batch_size, num_dims])
    return projection



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""API for world state representation."""

import abc
from typing import Any, List, Text, Tuple


class WorldState(abc.ABC):
  """Base class for world state representation.

    We can implement this class for world state representations in both
    sequential and matrix games.

    Attributes:
      chance_policy: Policy of the chance node in the game tree.
  """

  def __init__(self):
    self.chance_policy = {0: 1.0}
    self._history = []

  @abc.abstractmethod
  def get_distinct_actions(self) -> List[int]:
    """Returns all possible distinct actions in the game."""
    pass

  @abc.abstractmethod
  def is_terminal(self) -> bool:
    """Returns if the current state of the game is a terminal or not."""
    pass

  @abc.abstractmethod
  def get_actions(self) -> List[Any]:
    """Returns the list of legal actions from the current state of the game."""
    pass

  @abc.abstractmethod
  def get_infostate_string(self, player: int) -> Text:
    """Returns the string form of infostate representation of a given player.

    Args:
      player: Index of player.

    Returns:
      The string representation of the infostate of player.
    """

    pass

  @abc.abstractmethod
  def apply_actions(self, actions: Tuple[int, int, int]) -> None:
    """Applies the current player's action to change state of the world.

    At each timestep of the game, the state of the world is changing by the
    current player's action. At the same time, we should update self._history
    with actions, by appending actions to self._history.

    Args:
      actions: List of actions for chance node, player 1 and player 2.

    """
    pass

  @abc.abstractmethod
  def get_utility(self, player: int) -> float:
    """Returns player's utility when the game reaches to a terminal state.

    Args:
      player: Index of player.

    Returns:
      Utility that player receives when we reach a terminal state in the game.
    """
    pass





# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base classes for voting methods."""

import abc
from typing import Dict, List, NamedTuple, Tuple, Union
import numpy as np


# The id of an alternative can be a string or an integer.
AlternativeId = Union[str, int]

# List of alternative ids.
PreferenceList = List[AlternativeId]


# Basic type to represent a vote.
#    - The weight is an integer representing the number of voters
#    - The vote is a list of alternative ids, e.g. ["a", "b", "c"],
#      corresponding to a preference a > b > c.
class WeightedVote(NamedTuple):
  weight: int
  vote: PreferenceList


class PreferenceProfile(object):
  """Base class for preference profiles.

  IMPORTANT NOTE: see the assumptions below about indexing of alternatives.
  """
  _votes: List[WeightedVote]  # Tracks cast votes along with their count
  _alternatives_dict: Dict[AlternativeId, int]  # Maps ID to index
  # Identifiers for all possible alternatives
  _alternatives_ids: List[AlternativeId]

  def __init__(
      self,
      votes: Union[List[PreferenceList], List[WeightedVote], None] = None,
      alternatives: Union[List[AlternativeId], None] = None,
  ):
    """Initialize the preference profile.

    Args:
      votes: Either (i) a list of lists, each containing ids of alternatives,
        e.g. ["a", "b", "c"] signifiying a > b > c, or None for no votes, or
        (ii) a list of Vote tuples containing the weight and vote.
      alternatives: a list of alternatives ids.

    Note regarding how alternatives are indexed: if the second argument is
    passed, then the index of each alternative (e.g. when calling functions
    like margin_matrix etc.) will be assigned 0 up to the (number of
    alternatives) - 1 in the order of the list. If this argument is omitted,
    then alternatives will be indexed depending on when they are first seen
    (i.e. via a add_vote method) and so (only) in the latter case the indexing
    could depend on the order votes are added. Hence it is advised to pass in
    the list of alternatives to this function whenever they are known ahead of
    time.

    The alternatives_dict property below will return a dictionary of alternative
    IDs to index.
    """
    # List of Vote named tuples from above.
    self._votes: List[WeightedVote] = []
    # alternative id -> index (used for registering alternatives)
    self._alternatives_dict: Dict[AlternativeId, int] = {}
    # IDs (labels) of each alternative (usually strings). The alternative's
    # index is then the index of this array.
    self._alternatives_ids: List[AlternativeId] = []

    # Register the alternatives and add the votes, if any are provided.
    if alternatives is not None:
      for alternative in alternatives:
        self._register_alternative(alternative)
    if votes is not None:
      for vote in votes:
        self.add_vote(vote)
    if self._votes and not self._alternatives_ids:
      self._register_alternatives_from_votes()

  def _register_index_based_alternatives(self, num: int):
    """Register indices up to num-1 as possible alternatives."""
    for idx in range(num):
      self._register_alternative(idx)

  def _register_alternative(self, alternative: AlternativeId):
    """Add this alternative to internal records if not already there."""
    idx = self._alternatives_dict.get(alternative)
    if idx is None:
      self._alternatives_ids.append(alternative)
      self._alternatives_dict[alternative] = len(self._alternatives_ids) - 1
    assert (self._alternatives_ids[self._alternatives_dict[alternative]]
            == alternative)

  def _register_alternatives_from_votes(self):
    for vote in self._votes:
      for alternative in vote:
        self._register_alternative(alternative)

  def add_vote(
      self, vote: Union[PreferenceList, WeightedVote], weight: int = 1
  ):
    """Add a vote to this preference profile.

    Args:
      vote: Either (i) a list of ids, e.g. ["a", "b", "c"] signifying a > b > c,
          or, (ii) a Vote tuple containing both the weight and the vote of the
          form in (i).
      weight: the count, i.e. how many people have submitted this vote. Only
          used when the first argument is a list.
    """
    # For now support only integral weights (counts). Makes some things easier,
    # like N(x,y) and the margin matrices can be integers. Should be easy to
    # extend if we need to.
    assert isinstance(weight, int)
    assert weight > 0
    if isinstance(vote, WeightedVote):
      self._votes.append(vote)
      for alternative in vote.vote:
        self._register_alternative(alternative)
    else:
      weighted_vote = WeightedVote(weight, vote)
      self._votes.append(weighted_vote)
      for alternative in vote:
        self._register_alternative(alternative)

  def add_vote_from_values(
      self,
      values: Union[List[float], List[int]],
      tie_tolerance: float = 1e-10,
      weight: int = 1,
  ):
    """Adds a vote from a list of values.

    Note: this list is expected to cover all of the alternatives.

    WARNING: to ensure that ties are broken randomly, small random values are
    added to the values (within [0, tie_tolarance]). If the values are smaller
    than the tie_tolerance, this can be disabled by setting the tie_tolerance to
    0.

    Does not add the vote if the values are all within tie_tolerance of each
    other. For all others, adds a uniform * tie_tolerance to break ties.

    If the alternatives ids are not registered for this profile yet, then this
    method uses the indices of these values as the alternative IDs. Otherwise,
    the length of the array must be equal to the number of alternatives.

    Args:
      values: a list or numpy array of values for the alternative labeled by
          the index.
      tie_tolerance: a numerical threshold for determining ties.
      weight: the weight for the resulting vote.
    """
    # Check if any alternatives are registered for this profile. If not, then
    # first register ids for them all first.
    if not self._alternatives_ids:
      self._register_index_based_alternatives(len(values))
    else:
      assert len(values) == len(self._alternatives_ids)
    vals_copy = np.copy(np.asarray(values))
    max_val = vals_copy.max()
    min_val = vals_copy.min()
    if (max_val - min_val) < tie_tolerance:
      print(f"Warning: not casting vote from values: {vals_copy}")
      return
    # Add noise for tie_breaking
    vals_copy += tie_tolerance * np.random.uniform(size=len(vals_copy))
    vote = np.argsort(-vals_copy)
    # The vote is currently based on indices. Now convert to names.
    alternatives = self.alternatives
    assert alternatives
    assert len(alternatives) == len(vote)
    named_vote = []
    for idx in vote:
      assert 0 <= idx < len(alternatives)
      named_vote.append(alternatives[idx])
    self.add_vote(named_vote, weight=weight)

  @property
  def votes(self) -> List[WeightedVote]:
    """Returns a list of votes."""
    return self._votes

  @property
  def alternatives(self) -> List[AlternativeId]:
    """Returns a list of alternatives."""
    return self._alternatives_ids

  @property
  def alternatives_dict(self) -> Dict[AlternativeId, int]:
    """Returns a dict of alternative id -> index for each alternative."""
    return self._alternatives_dict

  def num_alternatives(self) -> int:
    return len(self._alternatives_ids)

  def num_votes(self) -> int:
    """Returns the number of votes."""
    total = 0
    for vote in self._votes:
      total += vote.weight
    return total

  def pref_matrix(self) -> np.ndarray:
    """Returns the candidate preference matrix for this profile.

    Define N(x,y) as number of voters that prefer x > y. The candidate
    preference matrix is one whose entries are N(x,y) for row x and column y.
    """
    # First map the alternatives to indices.
    m = self.num_alternatives()
    mat = np.zeros(shape=(m, m), dtype=np.int32)
    for vote in self._votes:
      vote_len = len(vote.vote)
      for i in range(vote_len):
        for j in range(i + 1, vote_len):
          # vote.vote[i] > vote.vote[j]
          idx_i = self._alternatives_dict[vote.vote[i]]
          idx_j = self._alternatives_dict[vote.vote[j]]
          mat[idx_i, idx_j] += vote.weight
    return mat

  def pairwise_count_matrix(self) -> np.ndarray:
    """Returns a matrix of pairwise for this profile.

    For each entry (i,j) in the matrix, entry[i,j] is the number of votes in
    the profile that contain both alternative i and alternative j.
    """
    pref_matrix = self.pref_matrix()
    return pref_matrix + pref_matrix.T

  def margin_matrix(self) -> np.ndarray:
    """Returns the margin matrix for this profile.

    Define N(x,y) = number of voters that prefer x > y. The margin matrix
    is a num_alternatives x num_alternatives whose entry at (r,c) is:
    delta(r,c) = N(r, c) - N(c, r). The r and c refer to columns, which
    correspond to the indices in the list returned by self.alternatives.
    """
    pref_matrix = self.pref_matrix()
    return pref_matrix - pref_matrix.T

  def condorcet_winner(
      self, strong: bool = True, margin_matrix: Union[np.ndarray, None] = None
  ):
    """Returns the Condorcet winner(s).

    Args:
      strong: whether it's a strong Condorcet winner (see below).
      margin_matrix: the margin matrix (optional: only used to to avoid
          recomputing).

    Returns:
      A list containing the Condorcet winners. There may be multiple weak
      Condorcet winners, but there is at most one strong winner.

    A strong Condorcet winner is an alternative a* in A such that for all
    a' in A: N(a*, a') > N(a', a*). A weak Condorcet winner is a similar
    definition using great-than-or-equal-to >=.
    """
    condorcet_winners = []
    if margin_matrix is None:
      margin_matrix = self.margin_matrix()
    for alt_idx in range(self.num_alternatives()):
      if strong and np.all(np.delete(margin_matrix[alt_idx] > 0, alt_idx)):
        # Don't count the diagonal 0 in the checking of > 0.
        condorcet_winners.append(self._alternatives_ids[alt_idx])
      elif not strong and np.all(margin_matrix[alt_idx] >= 0):
        condorcet_winners.append(self._alternatives_ids[alt_idx])
    if strong:
      assert len(condorcet_winners) <= 1
    return condorcet_winners

  def group(self):
    """Group up the votes.

    This will combine multiple identical votes into the smallest set of unique
    weighted votes.
    """
    old_votes = self._votes
    self._votes = []
    while old_votes:
      vote = old_votes[0].vote
      total_weight = old_votes[0].weight
      del old_votes[0]
      i = 0
      while i < len(old_votes):
        if old_votes[i].vote == vote:
          total_weight += old_votes[i].weight
          del old_votes[i]
        else:
          i += 1
      self._votes.append(WeightedVote(total_weight, vote))

  def ungroup(self):
    """Splits the votes into individual votes (each with weight of 1)."""
    old_votes = self._votes
    self._votes = []
    for vote in old_votes:
      for _ in range(vote.weight):
        self._votes.append(WeightedVote(1, vote.vote))

  def __str__(self) -> str:
    """Get a string representation of this profile."""
    string = ""
    for vote in self._votes:
      string += str(vote) + "\n"
    return string

  def total_weight(self) -> int:
    w = 0
    for vote in self._votes:
      w += vote.weight
    return w

  def get_weight(self, vote: PreferenceList) -> int:
    total_weight = 0
    for v in self._votes:
      if v.vote == vote:
        total_weight += v.weight
    return total_weight

  def set_weight(self, index: int, value: int):
    self._votes[index] = self._votes[index]._replace(weight=value)

  def set_all_weights(self, value: int):
    """Sets the weight of all the votes to the specified value."""
    for i in range(len(self._votes)):
      self.set_weight(i, value)

  def to_list_of_tuples(
      self, convert_alternatives_to_strings: bool = False
  ) -> List[Tuple[float, PreferenceList]]:
    """Returns a list of (alternative, score) tuples."""
    list_of_tuples = []
    for vote in self._votes:
      vote_lst = (
          [str(a) for a in vote.vote]
          if convert_alternatives_to_strings
          else vote.vote
      )
      list_of_tuples.append((vote.weight, vote_lst))
    return list_of_tuples


class RankOutcome(object):
  """Basic object for outcomes of the voting methods."""

  def __init__(self, rankings=None, scores=None):
    self._rankings: List[AlternativeId] = rankings
    self._scores: List[float] = scores
    self._rank_dict: Dict[AlternativeId, int] = None
    if self._rankings is not None:
      self.make_rank_dict()

  def unpack_from(
      self, ranked_alternatives_and_scores: List[Tuple[AlternativeId, float]]
  ):
    """A rank outcome that comes packed as (alternative id, score) tuples."""
    self._rankings, self._scores = zip(*ranked_alternatives_and_scores)
    self._rankings = list(self._rankings)
    self._scores = list(self._scores)
    self.make_rank_dict()

  @property
  def ranking(self) -> List[AlternativeId]:
    """Returns an ordered list W of alternatives' ids (winner is first)."""
    return self._rankings

  @property
  def scores(self) -> List[float]:
    """Returns a alternative's scores S (in the same order as the ranking)."""
    return self._scores

  def ranking_with_scores(self) -> Tuple[List[AlternativeId], List[float]]:
    """Returns an ordered list of alternative ids and dict of scores W, S."""
    return self._rankings, self._scores

  def make_rank_dict(self):
    """Makes the rank dictionary from the rankings and scores."""
    self._rank_dict = {}
    for r, alt in enumerate(self._rankings):
      self._rank_dict[alt] = r

  def get_rank(self, alternative: AlternativeId) -> int:
    """Returns the rank of a specific alternative."""
    return self._rank_dict[alternative]

  def get_score(self, alternative: AlternativeId) -> float:
    """Returns the score of a specific alternative."""
    return self._scores[self.get_index(alternative)]

  def get_index(self, alternative: AlternativeId) -> int:
    """Returns the index of a specific alternative."""
    return self._rankings.index(alternative)

  def __str__(self) -> str:
    str_rep = "Rank: " + str(self._rankings) + "\n"
    if self._scores is not None:
      str_rep += "Scores: " + str(self._scores)
    return str_rep

  def pretty_table_string(self, top: Union[int, None] = None):
    """Return an easier-to-read table for the rankings and scores.

    Args:
      top: (optional) if specified, only returns the top `top` alternatives.

    Returns:
      An easier-to-read table string.
    """
    if top is None:
      top = len(self._rankings)
    max_len = -1
    for i, alt in enumerate(self._rankings):
      if i == top:
        break
      max_len = max(max_len, len(str(alt)))
    table_string = ""
    max_len += 1
    for i, alt in enumerate(self._rankings):
      if i == top:
        break
      score = self._scores[i]
      prefix = f"    Rank {i+1}: "
      while len(prefix) < 14:
        prefix += " "
      prefix += str(alt)
      while len(prefix) < (14 + max_len):
        prefix += " "
      table_string += f"{prefix} ({score})\n"
    return table_string

  def pretty_latex_table(
      self, header: Union[str, None] = None, top: Union[int, None] = None
  ):
    """Return an easier-to-read table string for the rankings and scores.

    The string returned include LaTeX formatting for putting the tables into
    papers.

    Args:
      header: (optional) if specified, uses this as the header of the table.
      top: (optional) if specified, only returns the top `top` alternatives.

    Returns:
      An easier-to-read table string (with LaTeX formattinf)
    """

    if top is None:
      top = len(self._rankings)
    table_string = "\\begin{center}\n\\begin{tabular}{|c|ll|}\n"
    if header is not None:
      table_string += "\\multicolumn{3}{c}{\\bf " + header + "}\\\\\n\\hline\n"
    table_string += "Rank & Agent & Score\\\\\n\\hline\n"
    for i, alt in enumerate(self._rankings):
      if i == top:
        break
      score = self._scores[i]
      # table_string += f"{i+1} & \\textsc" + "{"
      table_string += f"{i+1} & " + "{\\tt "
      table_string += f"{alt}" + "} & " + f"{score}\\\\\n"
    table_string += "\\hline\n"
    table_string += "\\end{tabular}\n\\end{center}"
    return table_string


class AbstractVotingMethod(metaclass=abc.ABCMeta):
  """Abstract base class for voting methods."""

  @abc.abstractmethod
  def __init__(self, **method_specific_kwargs):
    """Initializes the voting method.

    Args:
      **method_specific_kwargs: optional extra args.
    """

  @abc.abstractmethod
  def name(self) -> str:
    """Returns the name of the voting method."""

  @abc.abstractmethod
  def run_election(self, profile: PreferenceProfile) -> RankOutcome:
    """Runs the election and returns the result.

    Args:
        profile: a preference profile.

    Returns:
        a RankOutcome object that can be queried for the results.
    """

  def is_valid_profile(self, profile: PreferenceProfile) -> bool:
    """Returns true if a profile is valid.
 
    A valid profile is valid if it contains at least one vote and one
    alternative. Most voting schemes can't run unless the profile is valid.

    Args:
      profile: the profile to check.
    """
    return profile.num_votes() > 0 and profile.num_alternatives() > 0



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for running tests."""

import dataclasses
import enum

from typing import List

import numpy as np

from open_spiel.python.games.chat_games.envs.comm_substrates import emails


class TestLLM(enum.Enum):
  MOCK = 0


@dataclasses.dataclass(frozen=True)
class MockScore:
  logprob: float


class MockModel():
  """Mock LLM model."""

  def __init__(self, name):
    self.name = name


class MockResponse():
  """Mock LLM response."""

  def __init__(self, text):
    self.text = text


class MockClient():
  """Mock LLM client."""

  def __init__(self):
    # for cycling through mock response options
    self._idxs = {'names': 0,
                  'tones': 0,
                  'examples': 0}

  def sample(self, model: str, length: int, seed: int, prompt: str
             ) -> MockResponse:
    """Returns string responses according to fixed prompt styles."""
    del model, length, seed
    prompt_lower = prompt.lower()
    if 'names' in prompt_lower:
      dummy_names = ['Suzy', 'Bob', 'Alice', 'Doug', 'Arun', 'Maria', 'Zhang']
      dummy_name = dummy_names[self._idxs['names']]
      self._idxs['names'] = (self._idxs['names'] + 1) % len(dummy_names)
      return MockResponse(dummy_name + '\n')
    elif 'tones' in prompt_lower:
      dummy_tones = ['Happy', 'Sad', 'Angry']
      dummy_tone = dummy_tones[self._idxs['tones']]
      self._idxs['tones'] = (self._idxs['tones'] + 1) % len(dummy_tones)
      return MockResponse(dummy_tone + '\n')
    elif 'list of items' in prompt_lower:
      num_examples = 10
      dummy_examples = [f'Example-{i}' for i in range(num_examples)]
      dummy_example = dummy_examples[self._idxs['examples']]
      self._idxs['examples'] = (self._idxs['examples'] + 1) % num_examples
      return MockResponse(dummy_example + '\n')
    elif 'score' in prompt_lower or 'value' in prompt_lower:
      return MockResponse('5\n')
    elif 'summary' in prompt_lower:
      return MockResponse('This is a summary of the dialogue. We are happy.\n')
    elif emails.BLOCK_OPT in prompt:
      return MockResponse('\nThat all sounds good to me.\n')
    else:
      raise ValueError('Prompt not recognized!\n\n' + prompt)

  def score(self, model: str, prompt: str) -> List[MockScore]:
    del model, prompt
    return [MockScore(logprob=-1)]

  def list_models(self) -> List[MockModel]:
    dummy_models = ['dummy_model']
    models = [MockModel(model_name) for model_name in dummy_models]
    return models


class MockLLM():
  """Mock LLM."""

  def __init__(self):
    self.client = MockClient()
    self.model = 'dummy_model'

  def generate_response(self, prompt: str, seed: int,
                        num_output_tokens: int) -> str:
    response = self.client.sample(
        model=self.model,
        length=num_output_tokens,
        seed=seed,
        prompt=prompt
        )
    return response.text

  def generate_bool(self, prompt: str, seed: int) -> bool:
    del seed
    score_true = self.client.score(model=self.model, prompt=prompt + 'Yes')
    score_false = self.client.score(model=self.model, prompt=prompt + 'No')
    if score_true > score_false:
      return True
    else:
      return False


class MockTokenizer():
  """Mock Tokenizer."""

  def to_int(self, text: str) -> np.ndarray:
    return np.zeros(len(text), dtype=np.int32)


class MockVectorizer():
  """Mock Vectorizer."""

  def __init__(self):
    self.tokenizer = MockTokenizer()

  def vectorize(self, text: str, obs_size: int) -> np.ndarray:
    observation = self.tokenizer.to_int(text)[:obs_size]
    num_pad = max(0, obs_size - observation.size)
    observation = np.pad(observation, (0, num_pad))
    return observation



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Counterfactual regret minimization (CFR) experiment.

Runs OpenSpiel CFR on a chat game.
"""

import dataclasses
import enum

from typing import Callable, Union

from absl import app
from absl import flags
from absl import logging

import ml_collections

import numpy as np

from open_spiel.python import policy as pyspiel_policy
from open_spiel.python.algorithms import expected_game_score

from open_spiel.python.games import chat_game  # pylint: disable=unused-import
from open_spiel.python.games.chat_games import chat_game_base

from open_spiel.python.games.chat_games.configs import config_debate
from open_spiel.python.games.chat_games.configs import config_schedule_meeting_w_dow
from open_spiel.python.games.chat_games.configs import config_schedule_meeting_w_tone
from open_spiel.python.games.chat_games.configs import config_trade_fruit_w_tone

from open_spiel.python.games.chat_games.envs.comm_substrates import schedules

from open_spiel.python.games.chat_games.utils import test_utils as chat_test_utils

import pyspiel


_SAVE_PATH = flags.DEFINE_string("save_path",
                                 default="",
                                 help="path for writing results")

LLM_TYPE = chat_test_utils.TestLLM.MOCK


class Domain(enum.StrEnum):
  TRADE_FRUIT_W_TONE = enum.auto()
  DEBATE_W_STYLE = enum.auto()
  SCHEDULE_MEETING_W_DOW = enum.auto()
  SCHEDULE_MEETING_W_TONE = enum.auto()


def new_debate_scenario_config(
    config: ml_collections.config_dict.ConfigDict,
    game_id: int,
) -> ml_collections.config_dict.ConfigDict:
  """Creates a new debate scenario config with a new topic.

  Arguments:
    config: the original debate scenario config dict (this should contain
      examples for generating new scenarios)
    game_id: int, will index into set of 20 debate topics found in
      https://www.englishclub.com/speaking/agreeing-disagreeing-topics.php
  Returns:
    new_config: debate config with redefined debate topic
  """
  # https://www.englishclub.com/speaking/agreeing-disagreeing-topics.php
  topics = ["Breakfast is the most important meal of the day.",
            "Swimming in the ocean is better than swimming in a public pool.",
            "Alcohol should be illegal.",
            "Children should provide room and board for their aging parents.",
            "Studying grammar is more important than practising conversation " +
            "skills.",
            "Television is the leading cause of violence in todays society.",
            "Dogs make better companions than cats.",
            "Smoking should be permitted in public places.",
            "Females are better students than males.",
            "A parent shouldn't pierce a babys ears.",
            "Women should be allowed to go topless in public.",
            "Lawyers should make a higher salary than nurses.",
            "Everyone should plan their own funeral.",
            "Reading English is more difficult than writing English.",
            "Summer is the best season of the year.",
            "Children under 13 should not be allowed to babysit.",
            "High school students should wear uniforms.",
            "21 should be the legal driving age around the world.",
            "Rock and Roll is the best kind of music.",
            "The government should pay for post secondary education."]

  topic = topics[game_id]
  config.game.given_private_info["topic"] = [topic, topic]

  return config


def same_scenario_config(
    config: ml_collections.config_dict.ConfigDict,
    game_id: int,
) -> ml_collections.config_dict.ConfigDict:
  """Dummy function for games that don't need any config modification.

  Arguments:
    config: the original game scenario config dict (this should contain
      examples for generating new scenarios)
    game_id: int, unused
  Returns:
    new_config: original game config
  """
  del game_id

  return config


def get_config_debate(config: ml_collections.config_dict.ConfigDict):
  """Get config for imitation dataset construction of debates."""

  config.config_rnd = config_debate.get_config()
  config.new_config = new_debate_scenario_config

  return config


def get_config_trade_fruit_w_tone(
    config: ml_collections.config_dict.ConfigDict,
):
  """Get config for imitation dataset construction of trading fruit."""

  config.config_rnd = config_trade_fruit_w_tone.get_config()
  config.new_config = same_scenario_config

  return config


def get_config_schedule_meeting_w_dow(
    config: ml_collections.config_dict.ConfigDict,
):
  """Get config for imitation dataset construction of meeting scheduling dow."""

  config.config_rnd = config_schedule_meeting_w_dow.get_config()
  config.new_config = same_scenario_config

  return config


def get_config_schedule_meeting_w_tone(
    config: ml_collections.config_dict.ConfigDict,
):
  """Get config for imitation dataset construction of meeting scheduling dow."""

  config.config_rnd = config_schedule_meeting_w_tone.get_config()
  config.new_config = same_scenario_config

  return config


def get_config():
  """Get configuration for imitation dataset construction."""
  config = ml_collections.config_dict.ConfigDict()

  config.game_string = "chat_game"
  config.game_id = 0
  config.seed = 34239871
  config.num_demos = 10
  config.num_iters = 4
  config.domain = Domain.SCHEDULE_MEETING_W_TONE

  if config.domain == Domain.DEBATE_W_STYLE:
    config = get_config_debate(config)
  elif config.domain == Domain.TRADE_FRUIT_W_TONE:
    config = get_config_trade_fruit_w_tone(config)
  elif config.domain == Domain.SCHEDULE_MEETING_W_DOW:
    config = get_config_schedule_meeting_w_dow(config)
    config.substrate = schedules
  elif config.domain == Domain.SCHEDULE_MEETING_W_TONE:
    config = get_config_schedule_meeting_w_tone(config)
  else:
    raise ValueError("Unknown domain: %s" % config.domain)

  return config


@dataclasses.dataclass(frozen=True)
class InfoStateRecord:
  observation: str | np.ndarray
  observation_str: str
  probabilities: list[float]
  actions: list[int]
  prev_message: str
  prev_speaker: int
  prev_action_strs: list[str]


@dataclasses.dataclass(frozen=False)
class GameStats:
  num_states: int = 0
  num_chance_nodes: int = 0
  num_decision_nodes: int = 0
  num_simultaneous_nodes: int = 0
  num_terminals: int = 0
  info_state_dict: dict[str, InfoStateRecord] = dataclasses.field(
      default_factory=dict)


@dataclasses.dataclass(frozen=True)
class EqRecord:
  nash_conv: float
  payoffs_eq_vs_bg_any: list[float]
  payoffs_any: list[float]
  payoffs_eq: list[float]


def record_info_state_data(
    state: pyspiel.State,
    policy: pyspiel.Policy,
    observer: Union[None, chat_game_base.ChatGameObserverBase] = None,
    vectorize: Union[None, Callable[[str, int], np.ndarray]] = None,
) -> InfoStateRecord:
  """Return observation and equilibrium strategy for a given state+policy."""
  pi = policy.action_probabilities(state)
  action_list = list(pi.keys())
  prob_list = list(pi.values())
  if observer is not None:
    info_str = observer.string_from(state, player=state.current_player())
    if vectorize is not None:
      info = vectorize(info_str, 768)
    else:
      info = info_str
  else:
    info = info_str = str(state)
  prev_msg = ""
  prev_speaker = -1
  prev_action_strs = []
  if state.played_actions:
    prev_action = state.played_actions[-1]
    prev_msg = state.dialogue[-1]
    prev_speaker = state.speakers[-1]
    prev_speaker = int(prev_speaker)
    prev_action_dict = state.unravel_flat_action_to_dict(prev_speaker,
                                                         prev_action)
    action_keys = state.prompt_actions.keys()
    prev_action_strs = [prev_action_dict["action"][key] for key in action_keys]
  sample = InfoStateRecord(info, info_str, prob_list, action_list,
                           prev_msg, prev_speaker, prev_action_strs)
  return sample


# traverses game tree and records game stats like info states.
def traverse_game_tree(
    game: pyspiel.Game,
    state: pyspiel.State,
    game_stats: GameStats,
    policy: pyspiel.Policy,
    observer: Union[None, chat_game_base.ChatGameObserverBase] = None,
    vectorize: Union[None, Callable[[str, int], np.ndarray]] = None,
):
  """Traverse the game tree and record GameStats in place.

  Args:
    game: pyspiel.Game
    state: initial pyspiel.State
    game_stats: empty GameStats object
    policy: pyspiel Policy
    observer: pyspiel Observer
    vectorize: method to vectorize a string
  """
  if state.is_terminal():
    game_stats.num_terminals += 1
  elif state.is_chance_node():
    game_stats.num_chance_nodes += 1
    for outcome in state.legal_actions():
      child = state.child(outcome)
      traverse_game_tree(game, child, game_stats, policy, observer, vectorize)
  elif state.is_simultaneous_node():
    game_stats.num_simultaneous_nodes += 1
    # TODO(imgemp): need to implement recording data for simultaneous
    # Using joint actions for convenience. Can use legal_actions(player) to
    # and state.apply_actions when walking over individual players
    for joint_action in state.legal_actions():
      child = state.child(joint_action)
      traverse_game_tree(game, child, game_stats, policy, observer, vectorize)
  else:
    game_stats.num_decision_nodes += 1
    if game.get_type().provides_information_state_string:
      sample = record_info_state_data(state, policy, observer, vectorize)
      game_stats.info_state_dict[
          state.information_state_string()] = sample
    for outcome in state.legal_actions():
      child = state.child(outcome)
      traverse_game_tree(game, child, game_stats, policy, observer, vectorize)


class ImitationDatasetConstructor():
  """Construct a dataset of (observation, CFR strategy) for imitation."""

  def __init__(self, save_path, config):
    self.save_path = save_path
    self.game_string = config.game_string
    self.game_id = config.game_id
    self.seed = config.seed
    self.num_demos = config.num_demos
    self.num_iters = config.num_iters
    self.domain = config.domain.value
    self.config_rnd = config.config_rnd
    self.new_config = config.new_config

    self._rnd = np.random.RandomState(self.seed)

    self.reporting = ImitationDatasetConstructorReporting(
        save_path=self.save_path,
        experiment_name="imitation_dataset_construction",
        game_string=self.game_string,
        game_id=self.game_id,
        seed=self.seed,
        num_demos=self.num_demos,
        num_iters=self.num_iters,
        domain=self.domain)

  def sample_to_dict(
      self,
      info_state_string: str,
      sample: InfoStateRecord,
      eq_record: EqRecord):
    """Constructs a dict mapping named keys to values in arguments."""

    sample_dict = {}
    sample_dict["info_state_string"] = info_state_string
    sample_dict["observation"] = sample.observation
    sample_dict["observation_str"] = sample.observation_str
    sample_dict["probabilities"] = sample.probabilities
    sample_dict["actions"] = sample.actions
    sample_dict["prev_message"] = sample.prev_message
    sample_dict["prev_speaker"] = sample.prev_speaker
    sample_dict["prev_action_strs"] = sample.prev_action_strs
    sample_dict["nash_conv"] = eq_record.nash_conv
    sample_dict["payoffs_eq_vs_bg_any"] = eq_record.payoffs_eq_vs_bg_any
    sample_dict["payoffs_any"] = eq_record.payoffs_any
    sample_dict["payoffs_eq"] = eq_record.payoffs_eq
    return sample_dict

  def eval_vs_any(self, game: pyspiel.Game, eq: pyspiel.Policy
                  ) -> EqRecord:
    """Evaluates the equilibrium against a background 'any' policy.

    Arguments:
      game: pyspiel.Game
      eq: pyspiel.Policy equilibrium policy (e.g., result of CFR)
    Returns:
      EqRecord containing
        ne_conv: float, sum of gains from each player best responding to eq
        payoffs_eq_vs_bg_any: list of floats, payoffs for each player when
          playing their side of equilibrium against background agents that all
          play 'any'
        payoff_any: list of floats, payoffs for each player when everyone plays
          'any' policy
        payoff_eq: list of floats, payoffs for each player when everyone plays
          equilibrium policy
    """
    ne_conv = pyspiel.nash_conv(game, eq)

    # construct pyspiel.Policy to play "any" tone (null strategy)
    # the action set is assumed to be (msg_receiver, prompt_action)
    # and "any" is assumed to be the last action in the prompt_action_list
    num_players = game.num_players()
    num_prompt_actions = game.num_distinct_actions() // num_players
    payoffs_eq_vs_bg_any = []
    one_hot_any = [0.0 for _ in range(game.num_distinct_actions())]
    for i in range(num_players):
      idx = i * num_prompt_actions + (num_prompt_actions - 1)
      one_hot_any[idx] = 1 / float(num_players)
    policy_any = dict(zip(range(len(one_hot_any)), one_hot_any))

    def callable_policy(state):
      del state
      return policy_any  # pylint:disable=cell-var-from-loop

    # compute expected payoffs for each player playing eq against "any" bg strat
    for i in range(num_players):
      policies = []
      for j in range(num_players):
        if i == j:
          # grab player i's side of avg_policy (eq_i)
          eq_i = pyspiel_policy.pyspiel_policy_to_python_policy(game,
                                                                eq,
                                                                players=[i])
          policies.append(eq_i)
        else:
          # setting player j policy to "any"
          p_j = pyspiel_policy.tabular_policy_from_callable(game,
                                                            callable_policy,
                                                            players=[j])
          policies.append(p_j)
      state = game.new_initial_state()
      payoff_array = expected_game_score.policy_value(state, policies)
      payoffs_eq_vs_bg_any.append(payoff_array[i])

    # compute expected payoffs when everyone plays "any" strategy
    policies = []
    for j in range(num_players):
      p_j = pyspiel_policy.tabular_policy_from_callable(game,
                                                        callable_policy,
                                                        players=[j])
      policies.append(p_j)
    state = game.new_initial_state()
    payoffs_any = expected_game_score.policy_value(state, policies)

    # compute expected payoffs when everyone plays eq strategy
    policies = []
    for j in range(num_players):
      # grab player j's side of avg_policy (eq_j)
      p_j = pyspiel_policy.pyspiel_policy_to_python_policy(game,
                                                           eq,
                                                           players=[j])
      policies.append(p_j)
    state = game.new_initial_state()
    payoffs_eq = expected_game_score.policy_value(state, policies)

    eq_record = EqRecord(ne_conv,
                         payoffs_eq_vs_bg_any,
                         payoffs_any,
                         payoffs_eq)

    return eq_record

  def construct_dataset(self):
    """Construct a dataset of (observation, optimal strategy) for imitation."""

    logging.info("Building vectorizer")
    vectorizer = chat_test_utils.MockVectorizer()
    vectorize = vectorizer.vectorize

    for demo in range(self.num_demos):
      logging.info("Creating new config for demo %d", demo)

      config = self.new_config(self.config_rnd, self.game_id)

      game = pyspiel.load_game(self.game_string, config.params.to_dict())

      seed = self._rnd.randint(42, 12345 + 1)
      game.load_chat_game(llm_type=LLM_TYPE,
                          vectorize=vectorize,
                          seed=seed,
                          **config.game)

      game_cached = pyspiel.convert_to_cached_tree(game)

      logging.info("Constructing CFR solver")
      cfr_solver = pyspiel.CFRSolver(game_cached)

      logging.info("Evaluating and Updating CFR policy")
      for i in range(self.num_iters):
        logging.info("CFR iteration %d", i)
        cfr_solver.evaluate_and_update_policy()

      logging.info("Averaging CFR policy")
      average_policy = cfr_solver.tabular_average_policy()

      eq_record = self.eval_vs_any(game_cached, average_policy)
      logging.info("NashConv: %f", eq_record.nash_conv)
      logging.info("Payoffs vs background any policy: %s",
                   eq_record.payoffs_eq_vs_bg_any)
      logging.info("Payoffs using any policy: %s", eq_record.payoffs_any)
      logging.info("Payoffs using eq policy: %s", eq_record.payoffs_eq)

      logging.info("Building info_state -> observation vectorizer")
      observer = game.make_py_observer()
      vectorizer = chat_test_utils.MockVectorizer()
      vectorize = vectorizer.vectorize

      logging.info("Traversing game tree and storing imitation policy")
      game_stats = GameStats()
      state = game.new_initial_state()
      traverse_game_tree(game, state, game_stats, average_policy,
                         observer=observer, vectorize=vectorize)
      h = f = "*" * 50
      for info_state_string in game_stats.info_state_dict:
        logging.info("%s\nInfo state string:\n%s\n%s", h, info_state_string, f)
        sample = game_stats.info_state_dict[info_state_string]
        results = self.sample_to_dict(info_state_string, sample, eq_record)
        self.reporting.report(demo, results)

      logging.info("Number of info states (length of policy): %d",
                   len(average_policy))


class ImitationDatasetConstructorReporting(object):
  """Utilities for logging an experiment run."""

  def __init__(
      self,
      save_path: str,
      experiment_name: str,
      game_string: str,
      game_id: int,
      seed: int,
      num_demos: int,
      num_iters: int,
      domain: str,
  ):
    self.save_path = save_path
    self.experiment_name = experiment_name
    self.game_string = game_string
    self.game_id = game_id
    self.seed = seed
    self.num_demos = num_demos
    self.num_iters = num_iters
    self.domain = domain

    config_dict_params = {}
    config_dict_params["experiment_name"] = self.experiment_name
    config_dict_params["game_string"] = self.game_string
    config_dict_params["seed"] = self.seed
    config_dict_params["num_demos"] = self.num_demos
    config_dict_params["num_iters"] = self.num_iters
    config_dict_params["domain"] = self.domain

    print("Config parameters:\n{:}".format(config_dict_params))

  def report(self, demo: int, results):
    """Report the exploitability."""
    print("CFR statistics ({:d}):\n{:}".format(demo, results))


def main(_):
  logging.set_verbosity(logging.ERROR)  # silence internal game logging
  save_path = _SAVE_PATH.value
  config = get_config()
  im = ImitationDatasetConstructor(save_path, config)
  im.construct_dataset()


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tensorflow.compat.v1 as tf

from open_spiel.python import rl_environment
from open_spiel.python.algorithms import nfsp

# Temporarily disable TF2 behavior until code is updated.
tf.disable_v2_behavior()


class NFSPTest(tf.test.TestCase):

  def test_run_kuhn(self):
    env = rl_environment.Environment("kuhn_poker")
    state_size = env.observation_spec()["info_state"][0]
    num_actions = env.action_spec()["num_actions"]

    with self.session() as sess:
      agents = [
          nfsp.NFSP(  # pylint: disable=g-complex-comprehension
              sess,
              player_id,
              state_representation_size=state_size,
              num_actions=num_actions,
              hidden_layers_sizes=[16],
              reservoir_buffer_capacity=10,
              anticipatory_param=0.1) for player_id in [0, 1]
      ]
      sess.run(tf.global_variables_initializer())

      for unused_ep in range(10):
        time_step = env.reset()
        while not time_step.last():
          current_player = time_step.observations["current_player"]
          current_agent = agents[current_player]
          agent_output = current_agent.step(time_step)
          time_step = env.step([agent_output.action])

        for agent in agents:
          agent.step(time_step)


class ReservoirBufferTest(tf.test.TestCase):

  def test_reservoir_buffer_add(self):
    reservoir_buffer = nfsp.ReservoirBuffer(reservoir_buffer_capacity=10)
    self.assertEmpty(reservoir_buffer)
    reservoir_buffer.add("entry1")
    self.assertLen(reservoir_buffer, 1)
    reservoir_buffer.add("entry2")
    self.assertLen(reservoir_buffer, 2)

    self.assertIn("entry1", reservoir_buffer)
    self.assertIn("entry2", reservoir_buffer)

  def test_reservoir_buffer_max_capacity(self):
    reservoir_buffer = nfsp.ReservoirBuffer(reservoir_buffer_capacity=2)
    reservoir_buffer.add("entry1")
    reservoir_buffer.add("entry2")
    reservoir_buffer.add("entry3")

    self.assertLen(reservoir_buffer, 2)

  def test_reservoir_buffer_sample(self):
    replay_buffer = nfsp.ReservoirBuffer(reservoir_buffer_capacity=3)
    replay_buffer.add("entry1")
    replay_buffer.add("entry2")
    replay_buffer.add("entry3")

    samples = replay_buffer.sample(3)

    self.assertIn("entry1", samples)
    self.assertIn("entry2", samples)
    self.assertIn("entry3", samples)


if __name__ == "__main__":
  tf.test.main()



# Copyright 2023 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Copeland's method.

Based on https://en.wikipedia.org/wiki/Copeland%27s_method.
"""

from open_spiel.python.voting import base


class CopelandVoting(base.AbstractVotingMethod):
  """Implements Copeland's method."""

  def __init__(self):
    pass

  def name(self) -> str:
    return "copeland"

  def run_election(self, profile: base.PreferenceProfile) -> base.RankOutcome:
    assert self.is_valid_profile(profile)
    copeland_scores = {}
    alternatives = profile.alternatives
    m = len(alternatives)
    margin_matrix = profile.margin_matrix()
    for r in range(m):
      alternative = alternatives[r]
      num_majority = (margin_matrix[r] > 0).sum()
      # Subtract one because we don't include the diagonal.
      num_ties = (margin_matrix[r] == 0).sum() - 1
      copeland_scores[alternative] = num_majority + 0.5 * num_ties
    sorted_scores = sorted(copeland_scores.items(), key=lambda item: item[1],
                           reverse=True)
    outcome = base.RankOutcome()
    outcome.unpack_from(sorted_scores)
    return outcome



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python dynamic routing game."""

from absl.testing import absltest

from open_spiel.python import games  # pylint:disable=unused-import
from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import cfr
from open_spiel.python.algorithms import expected_game_score
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import external_sampling_mccfr as external_mccfr
from open_spiel.python.algorithms import outcome_sampling_mccfr as outcome_mccfr
from open_spiel.python.games import dynamic_routing
from open_spiel.python.games import dynamic_routing_utils
import pyspiel

_NUM_ITERATION_CFR_TEST = 1


class DynamicRoutingGameTest(absltest.TestCase):

  def test_random_game(self):
    """Tests basic API functions with the standard game tests."""
    game = pyspiel.load_game("python_dynamic_routing")
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_game_as_turn_based(self):
    """Check the game can be converted to a turn-based game."""
    game = pyspiel.load_game("python_dynamic_routing")
    turn_based = pyspiel.convert_to_turn_based(game)
    pyspiel.random_sim_test(
        turn_based, num_sims=10, serialize=False, verbose=True)

  def test_game_as_turn_based_via_string(self):
    """Check the game can be created as a turn-based game from a string."""
    game = pyspiel.load_game(
        "turn_based_simultaneous_game(game=python_dynamic_routing())")
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_non_default_param_from_string(self):
    """Check params can be given through string loading."""
    game = pyspiel.load_game("python_dynamic_routing(max_num_time_step=5)")
    self.assertEqual(game.max_game_length(), 5)

  def test_non_default_param_from_dict(self):
    """Check params can be given through a dictionary."""
    game = pyspiel.load_game("python_dynamic_routing", {"max_num_time_step": 5})
    self.assertEqual(game.max_game_length(), 5)

  def test_action_consistency_convert_to_turn_based(self):
    """Check if the sequential game is consistent with the game."""
    game = pyspiel.load_game("python_dynamic_routing")
    seq_game = pyspiel.convert_to_turn_based(game)
    state = game.new_initial_state()
    seq_state = seq_game.new_initial_state()
    self.assertEqual(
        state.legal_actions(seq_state.current_player()),
        seq_state.legal_actions(),
        msg="The sequential actions are not correct.")

  def test_cfr_on_turn_based_game_with_exploitability(self):
    """Check if CFR can be applied to the sequential game."""
    game = pyspiel.load_game(
        "python_dynamic_routing(max_num_time_step=5,time_step_length=1.0)")
    seq_game = pyspiel.convert_to_turn_based(game)
    cfr_solver = cfr.CFRSolver(seq_game)
    for _ in range(_NUM_ITERATION_CFR_TEST):
      cfr_solver.evaluate_and_update_policy()
    exploitability.nash_conv(seq_game, cfr_solver.average_policy())

  def test_ext_mccfr_on_turn_based_game_with_exploitability(self):
    """Check if external sampling MCCFR can be applied."""
    game = pyspiel.load_game(
        "python_dynamic_routing(max_num_time_step=5,time_step_length=1.0)")
    seq_game = pyspiel.convert_to_turn_based(game)
    cfr_solver = external_mccfr.ExternalSamplingSolver(
        seq_game, external_mccfr.AverageType.SIMPLE)
    for _ in range(_NUM_ITERATION_CFR_TEST):
      cfr_solver.iteration()
    exploitability.nash_conv(seq_game, cfr_solver.average_policy())

  def test_int_mccfr_on_turn_based_game_with_exploitability(self):
    """Check if outcome sampling MCCFR can be applied."""
    game = pyspiel.load_game(
        "python_dynamic_routing(max_num_time_step=5,time_step_length=1.0)")
    seq_game = pyspiel.convert_to_turn_based(game)
    cfr_solver = outcome_mccfr.OutcomeSamplingSolver(seq_game)
    for _ in range(_NUM_ITERATION_CFR_TEST):
      cfr_solver.iteration()
    exploitability.nash_conv(seq_game, cfr_solver.average_policy())

  def test_creation_of_rl_environment(self):
    """Check if RL environment can be created."""
    game = pyspiel.load_game("python_dynamic_routing")
    seq_game = pyspiel.convert_to_turn_based(game)
    rl_environment.Environment(seq_game)

  def test_vehicle_origin_outside_network(self):
    """Check raise assertion if vehicle's origin is outside the Network."""
    vehicles = [dynamic_routing_utils.Vehicle("I->O", "D->E", 0)]
    with self.assertRaises(ValueError):
      dynamic_routing.DynamicRoutingGame(
          {
              "max_num_time_step": 10,
              "time_step_length": 0.5,
              "players": -1
          },
          vehicles=vehicles)

  def test_vehicle_destination_outside_network(self):
    """Check raise assertion if vehicle's destination is outside the Network."""
    vehicles = [dynamic_routing_utils.Vehicle("O->A", "E->F", 0)]
    with self.assertRaises(ValueError):
      dynamic_routing.DynamicRoutingGame(
          {
              "max_num_time_step": 10,
              "time_step_length": 0.5,
              "players": -1
          },
          vehicles=vehicles)

  def test_multiple_departure_time_vehicle(self):
    """Check that departure time can be define."""
    vehicles = [
        dynamic_routing_utils.Vehicle("O->A", "D->E", 0),
        dynamic_routing_utils.Vehicle("O->A", "D->E", 0.5),
        dynamic_routing_utils.Vehicle("O->A", "D->E", 1.0)
    ]
    game = dynamic_routing.DynamicRoutingGame(
        {
            "max_num_time_step": 10,
            "time_step_length": 0.5,
            "players": -1
        },
        vehicles=vehicles)
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_game_evolution_first_action_policy(self):
    """Check game deterministic evolution under first action policy."""
    # Test evolution of the game as expected (test value of the state).
    # test legal_actions().

  def test_observer_correct(self):
    """Check that the observer is correclty updated."""
    # Add test about observer and tensor being updated.

  def test_apply_actions_error_no_movement_with_negative_waiting_time(self):
    """Check that a vehicle cannot choose to not move if it has to move."""
    # Test apply_actions().

  def test_apply_actions_error_wrong_movement_with_negative_waiting_time(self):
    """Check that a vehicle cannot choose to move to a not successor link."""
    # Test apply_actions().

  def test_apply_actions_error_movement_with_positive_waiting_time(self):
    """Check that a vehicle cannot choose to move if it cannot move yet."""
    # Test apply_actions().

  def test_braess_paradox(self):
    """Test that Braess paradox can be reproduced with the mean field game."""
    num_player = 8
    braess_network = dynamic_routing_utils.Network(
        {
            "O": "A",
            "A": ["B", "C"],
            "B": ["C", "D"],
            "C": ["D"],
            "D": ["E"],
            "E": []
        },
        node_position={
            "O": (0, 0),
            "A": (1, 0),
            "B": (2, 1),
            "C": (2, -1),
            "D": (3, 0),
            "E": (4, 0)
        },
        bpr_a_coefficient={
            "O->A": 0,
            "A->B": 1.0,
            "A->C": 0,
            "B->C": 0,
            "B->D": 0,
            "C->D": 1.0,
            "D->E": 0
        },
        bpr_b_coefficient={
            "O->A": 1.0,
            "A->B": 1.0,
            "A->C": 1.0,
            "B->C": 1.0,
            "B->D": 1.0,
            "C->D": 1.0,
            "D->E": 1.0
        },
        capacity={
            "O->A": num_player,
            "A->B": num_player,
            "A->C": num_player,
            "B->C": num_player,
            "B->D": num_player,
            "C->D": num_player,
            "D->E": num_player
        },
        free_flow_travel_time={
            "O->A": 0,
            "A->B": 1.0,
            "A->C": 2.0,
            "B->C": 0.25,
            "B->D": 2.0,
            "C->D": 1.0,
            "D->E": 0
        })

    demand = [
        dynamic_routing_utils.Vehicle("O->A", "D->E") for _ in range(num_player)
    ]
    game = dynamic_routing.DynamicRoutingGame(
        {
            "time_step_length": 0.125,
            "max_num_time_step": 40
        },
        network=braess_network,
        vehicles=demand)

    class TruePathPolicy(policy.Policy):

      def __init__(self, game):
        super().__init__(game, list(range(num_player)))
        self._path = {}

      def action_probabilities(self, state, player_id=None):
        assert player_id is not None
        legal_actions = state.legal_actions(player_id)
        if not legal_actions:
          return {dynamic_routing_utils.NO_POSSIBLE_ACTION: 1.0}
        elif len(legal_actions) == 1:
          return {legal_actions[0]: 1.0}
        else:
          if legal_actions[0] == 1:
            if self._path[player_id] in ["top", "middle"]:
              return {1: 1.0}
            elif self._path[player_id] == "bottom":
              return {2: 1.0}
            else:
              raise ValueError()
          elif legal_actions[0] == 3:
            if self._path[player_id] == "top":
              return {4: 1.0}
            elif self._path[player_id] == "middle":
              return {3: 1.0}
            else:
              raise ValueError()
        raise ValueError(f"{legal_actions} is not correct.")

    class NashEquilibriumBraess(TruePathPolicy):

      def __init__(self, game):
        super().__init__(game)
        for player_id in range(num_player):
          if player_id % 2 == 0:
            self._path[player_id] = "middle"
          if player_id % 4 == 1:
            self._path[player_id] = "top"
          if player_id % 4 == 3:
            self._path[player_id] = "bottom"

    class SocialOptimumBraess(NashEquilibriumBraess):

      def __init__(self, game):
        super().__init__(game)
        for player_id in range(num_player):
          if player_id % 2 == 0:
            self._path[player_id] = "top"
          if player_id % 2 == 1:
            self._path[player_id] = "bottom"

    ne_policy = NashEquilibriumBraess(game)
    # Debug issue with nash conv computation and uncomment yhe following line.
    # self.assertEqual(exploitability.nash_conv(game, ne_policy), 0.0)
    self.assertSequenceAlmostEqual(
        -expected_game_score.policy_value(game.new_initial_state(), ne_policy),
        [3.75] * num_player)

    so_policy = SocialOptimumBraess(game)
    # Debug issue with nash conv computation and uncomment the following line.
    # self.assertEqual(exploitability.nash_conv(game, so_policy), 0.125)
    self.assertSequenceAlmostEqual(
        -expected_game_score.policy_value(game.new_initial_state(), so_policy),
        [3.5] * num_player)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Prompt-Space Response-Oracle (PSRO) experiment.

Runs PSRO exploring the space of `tones` with which to construct messages. Only
works with `tones` for now.
"""

import enum
import itertools
import math

from absl import app
from absl import flags
from absl import logging

import ml_collections

import nashpy
import numpy as np

from open_spiel.python.games import chat_game  # pylint: disable=unused-import
from open_spiel.python.games.chat_games.configs import config_schedule_meeting_w_tone
from open_spiel.python.games.chat_games.configs import config_trade_fruit_w_tone
from open_spiel.python.games.chat_games.envs.utils import text
from open_spiel.python.games.chat_games.utils import test_utils as chat_test_utils

import pyspiel


_SAVE_PATH = flags.DEFINE_string("save_path",
                                 default="",
                                 help="path for writing results")

LLM_TYPE = chat_test_utils.TestLLM.MOCK


class Domain(enum.StrEnum):
  TRADE_FRUIT_W_TONE = enum.auto()
  SCHEDULE_MEETING_W_TONE = enum.auto()


def get_config():
  """Get configuration for imitation dataset construction."""
  config = ml_collections.config_dict.ConfigDict()

  config.game_string = "chat_game"
  config.seed = 34239871
  config.num_iters = 4
  config.num_trials = 10
  config.num_candidates = 2
  config.domain = Domain.SCHEDULE_MEETING_W_TONE

  if config.domain == Domain.TRADE_FRUIT_W_TONE:
    config.env_config = config_trade_fruit_w_tone.get_config()
  elif config.domain == Domain.SCHEDULE_MEETING_W_TONE:
    config.env_config = config_schedule_meeting_w_tone.get_config()
  else:
    raise ValueError("Unknown domain: %s" % config.domain)

  return config


def sym(pt):
  """Symmetrize stack of payoff tensors (stacked along first dimension).

  A payoff tensor can be `symmetrized' by averaging over all possible
  permutations of the players. This means permuting the axes corresponding to
  the player strategies as well as the payoffs assigned to the players. E.g.,
  player A playing strategy 1 and player B playing strategy 3 is no different
  from player A playing strategy 3 and player B playing strategy 1 in a
  symmetric game. Note we permuted the strategies, but we must also permute the
  payoffs.

  Args:
    pt: tensor of shape: (num_players,) + (num_strategies,) * num_players
  Returns:
    pt_sym: symmetrized payoff tensor of same shape
  """
  num_players = len(pt.shape[1:])
  num_perms = math.factorial(num_players)
  pt_sym = np.zeros_like(pt)
  for _, perm_players in enumerate(itertools.permutations(range(num_players))):
    perm_axes = tuple([pi + 1 for pi in perm_players])
    permuted_tensor = np.transpose(pt, (0,) + perm_axes)[list(perm_players)]
    pt_sym += permuted_tensor / float(num_perms)
  return pt_sym


def random_policy(rnd, state):
  # all actions are legal for now
  rnd_action = tuple([rnd.choice(a) for a in state.num_actions])
  return np.ravel_multi_index(rnd_action, state.num_actions)


def fixed_prompt_policy(rnd, state, prompt_action_dict):
  # all actions are legal for now
  action = [rnd.choice(a) for a in state.num_actions]
  for prompt_key, prompt_action in prompt_action_dict.items():
    prompt_key_idx = 1 + state.header.action_keys.index(prompt_key)
    prompt_val_idx = state.prompt_actions[prompt_key].index(prompt_action)
    action[prompt_key_idx] = prompt_val_idx
  action = tuple(action)
  return np.ravel_multi_index(action, state.num_actions)


def mixed_prompt_policy(rnd, state, prompt_keys, mixture):
  # all actions are legal for now
  action = [rnd.choice(a) for a in state.num_actions]
  for prompt_key in prompt_keys:
    prompt_key_idx = 1 + state.header.action_keys.index(prompt_key)
    actions = state.prompt_actions[prompt_key]
    num_actions = len(actions)
    prompt_val_idx = rnd.choice(num_actions, p=mixture)
    action[prompt_key_idx] = prompt_val_idx
  action = tuple(action)
  return np.ravel_multi_index(action, state.num_actions)


def build_player_policy(policies):
  def player_policy(player_id, state):
    return policies[player_id](state)
  return player_policy


def simulate_dialogue(game, policy):
  """Simulate a dialogue and returns payoffs for each player."""

  state = game.new_initial_state()

  while not state.is_terminal():
    if state.is_chance_node():
      # Chance node: sample an outcome
      outcomes = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes)
      action = np.random.choice(action_list, p=prob_list)
      state.apply_action(action)
    else:
      # Decision node: sample action for the single current player
      action = policy(state.current_player(), state)
      state.apply_action(action)

  # Game is now done. Print utilities for each player
  returns = state.returns()

  return returns


def estimate_payoff_tensor(game, rnd, num_trials=5):
  """Simulate a batch of dialogues and returns payoffs for each player."""

  num_players = game.num_players()
  num_actions = len(game.given_prompt_actions["tone"])
  payoff_tensor = np.zeros(
      (num_trials, num_players) + (num_actions,) * num_players
  )

  joint_actions = list(itertools.product(range(num_actions),
                                         repeat=num_players))

  for trial in range(num_trials):
    for joint_action_idx in joint_actions:
      policies = []
      for _, tone_idx in zip(range(num_players), joint_action_idx):
        fixed_tone = {"tone": game.given_prompt_actions["tone"][tone_idx]}
        policy = lambda state: fixed_prompt_policy(rnd, state, fixed_tone)  # pylint:disable=cell-var-from-loop
        policies.append(policy)
      player_policy = build_player_policy(policies)

      returns = simulate_dialogue(game, player_policy)

      pt_index = (trial, slice(None)) + joint_action_idx

      payoff_tensor[pt_index] = returns

  return payoff_tensor


def score_candidate_responses(game_str, config, load_dict, rnd,
                              background_policies, candidates,
                              player_ids=(0,), num_trials=5):
  """Simulate a batch of dialogues and returns payoffs for each player."""

  num_players = config.params["num_players"]

  num_candidates = len(candidates)

  config.game.given_prompt_actions["tone"] += candidates
  num_actions = len(config.game.given_prompt_actions["tone"])
  config.params["num_distinct_actions"] = num_players * num_actions

  game = pyspiel.load_game(game_str, config.params.to_dict())

  game.load_chat_game(**load_dict, **config.game)

  payoffs = np.zeros((num_trials, len(player_ids), num_candidates))

  for player_id in player_ids:
    for trial in range(num_trials):
      for candidate_idx in range(num_candidates):
        policies = []
        for i in range(num_players):
          if player_id == i:
            fixed_tone = {"tone": candidates[candidate_idx]}
            policy = lambda state: fixed_prompt_policy(rnd, state, fixed_tone)  # pylint:disable=cell-var-from-loop
            policies.append(policy)
          else:
            policies.append(background_policies[i])
        player_policy = build_player_policy(policies)

        returns = simulate_dialogue(game, player_policy)

        payoffs[trial, player_id, candidate_idx] = returns[player_id]

  # undo changes to config (is this inplace?)
  config.game.given_prompt_actions["tone"] = config.game.given_prompt_actions[
      "tone"
  ][:-num_candidates]
  num_tones = len(config.game.given_prompt_actions["tone"])
  config.params["num_distinct_actions"] = num_players * num_tones

  return payoffs, candidates


def compute_sym_eq(pt):
  game = nashpy.Game(pt[0], pt[1])
  p1_traj, p2_traj = game.asymmetric_replicator_dynamics()
  p1_strat = np.mean(p1_traj, axis=0)
  p2_strat = np.mean(p2_traj, axis=0)
  return 0.5 * p1_strat + 0.5 * p2_strat


class PSRO():
  """Run prompt-space response oracle algorithm on chat game."""

  def __init__(self, save_path, config):
    self.save_path = save_path
    self.game_string = config.game_string
    self.seed = config.seed
    self.num_iters = config.num_iters
    self.num_trials = config.num_trials
    self.num_candidates = config.num_candidates
    self.domain = config.domain.value
    self.config = config.env_config

    self.rnd = np.random.RandomState(self.seed)

    self.num_players = self.config.params["num_players"]

    self.game = pyspiel.load_game(self.game_string,
                                  self.config.params.to_dict())

    vectorizer = chat_test_utils.MockVectorizer()
    vectorize = vectorizer.vectorize

    self.load_dict = {"llm_type": LLM_TYPE,
                      "vectorize": vectorize,
                      "seed": self.seed}

    self.game.load_chat_game(**self.load_dict, **self.config.game)

    self.reporting = PSROReporting(
        save_path=self.save_path,
        experiment_name="psro",
        game_string=self.game_string,
        seed=self.seed,
        num_iters=self.num_iters,
        num_trials=self.num_trials,
        num_candidates=self.num_candidates,
        domain=self.domain,
        base_candidates=list(self.config.game.given_prompt_actions["tone"]))

  def run(self):
    """Evaluate an imitation-learned policy."""

    for psro_iter in range(self.num_iters):

      pt = estimate_payoff_tensor(self.game,
                                  self.rnd,
                                  num_trials=self.num_trials)
      pt = pt.mean(axis=0)  # mean over trials
      pt = sym(pt)  # symmetrize the pt

      # compute eq
      sub_eq = compute_sym_eq(pt)  # assume symmetric ne

      # generate num_candidate tones
      actions = self.config.game.given_prompt_actions["tone"]
      candidates = self.game.generate_prompts("tone",
                                              actions,
                                              self.num_candidates,
                                              text.retrieve_alpha_block)
      new_actions = actions + candidates
      new_num_actions = len(new_actions)

      eq = np.zeros(new_num_actions) / float(new_num_actions)
      eq[:pt.shape[1]] = sub_eq

      background_policies = []
      for _ in range(self.num_players):
        bg_policy = lambda state: mixed_prompt_policy(self.rnd,
                                                      state,
                                                      ["tone"],
                                                      eq)  # pylint:disable=cell-var-from-loop
        background_policies.append(bg_policy)

      scores, candidates = score_candidate_responses(
          self.game_string,
          self.config,
          self.load_dict,
          self.rnd,
          background_policies,
          candidates,
          player_ids=(0,),
          num_trials=self.num_trials)

      mean_scores = np.mean(scores, axis=0)[0]  # only need player 0's scores
      br_idx = np.argmax(mean_scores)
      br = candidates[br_idx]

      self.config.game.given_prompt_actions["tone"] += [br]
      new_num_tones = len(self.config.game.given_prompt_actions["tone"])
      self.num_players = self.config.params["num_players"]
      new_num_distinct_actions = self.num_players * new_num_tones
      self.config.params["num_distinct_actions"] = new_num_distinct_actions

      self.game = pyspiel.load_game(self.game_string,
                                    self.config.params.to_dict())

      self.game.load_chat_game(**self.load_dict, **self.config.game)

      self.reporting.report(psro_iter,
                            pt,
                            br,
                            mean_scores,
                            candidates,
                            sub_eq)


class PSROReporting(object):
  """Utilities for logging an experiment run."""

  def __init__(self,
               save_path: str,
               experiment_name: str,
               game_string: str,
               seed: int,
               num_iters: int,
               num_trials: int,
               num_candidates: int,
               domain: str,
               base_candidates: list[str]):
    self.save_path = save_path
    self.experiment_name = experiment_name
    self.game_string = game_string
    self.seed = seed
    self.num_iters = num_iters
    self.num_trials = num_trials
    self.num_candidates = num_candidates
    self.domain = domain
    self.base_candidates = base_candidates

    config_dict_params = {}
    config_dict_params["experiment_name"] = self.experiment_name
    config_dict_params["game_string"] = self.game_string
    config_dict_params["seed"] = self.seed
    config_dict_params["num_iters"] = self.num_iters
    config_dict_params["num_trials"] = self.num_trials
    config_dict_params["num_candidates"] = self.num_candidates
    config_dict_params["domain"] = self.domain
    config_dict_params["base_candidates"] = self.base_candidates

    print("Config parameters:\n{:}".format(config_dict_params))

  def report(self,
             psro_iter: int,
             payoff_tensor: np.ndarray,
             br: str,
             mean_scores: np.ndarray,
             candidates: np.ndarray,
             eq: np.ndarray):
    """Report the psro statistics."""
    psro_stats_dict = {}
    psro_stats_dict["psro_iter"] = psro_iter
    psro_stats_dict["payoff_tensor"] = payoff_tensor
    psro_stats_dict["br"] = br
    psro_stats_dict["mean_scores"] = mean_scores
    psro_stats_dict["candidates"] = candidates
    psro_stats_dict["eq"] = eq

    print("PSRO statistics ({:d}):\n{:}".format(psro_iter, psro_stats_dict))


def main(_):
  logging.set_verbosity(logging.ERROR)  # silence internal game logging
  save_path = _SAVE_PATH.value
  config = get_config()
  psro = PSRO(save_path, config)
  psro.run()


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Optimization algorithms to compute (C)CE weights."""

import numpy as np
import scipy.optimize
import scipy.sparse.linalg


# pylint: disable=invalid-name
def get_proba_constraints_positivity(nus):
  A = np.zeros((nus.shape[0], 1 + nus.shape[0]))
  A[:, 1:] = -np.eye(nus.shape[0])
  return A, np.zeros(A.shape[0])


def get_proba_constraint_sum_eq(nus):
  A = np.ones((1, 1 + nus.shape[0]))
  A[0, 0] = 0.0
  return A, np.array([1.0])


def compress_internal_weights(nus, regrets, rewards, lbd=0.0):
  """Computes distribution over `nus` while minimizing internal regret.

  Args:
    nus: [T, P] array, T the number of different population distributions, P the
      number of different policies.
    regrets: [T, P, P] array, regrets[t, i, j] = payoff for switching from
      policy i to j at time t.
    rewards: [T, P] array, T the number of different population distributions, P
      the number of different policies
    lbd: Sparsity argument.

  Returns:
    Computed distribution over `nus`.
  """

  def get_c(nus):
    return np.concatenate(
        (np.array([1.0]), -lbd * np.sum(rewards * nus, axis=1))
    )

  def get_max_constraint(regrets):
    regrets = np.transpose(np.array(regrets), axes=[0, 2, 1])
    regrets = regrets.reshape(-1, regrets.shape[-1])
    A = np.zeros((regrets.shape[0], 1 + regrets.shape[1]))
    A[:, 1:] = regrets
    A[:, 0] = -1.0

    b = np.zeros(A.shape[0])
    return A, b

  def get_a_ub(nus, regrets):
    Amax, bmax = get_max_constraint(regrets)
    Apos, bpos = get_proba_constraints_positivity(nus)
    return np.concatenate((Amax, Apos), axis=0), np.concatenate(
        (bmax, bpos), axis=0
    )

  c = get_c(nus)

  A_ub, b_ub = get_a_ub(nus, regrets)
  A_eq, b_eq = get_proba_constraint_sum_eq(nus)

  res = scipy.optimize.linprog(
      c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, options={'tol': 1e-10}
  )
  new_weights = res.x
  return new_weights[1:]


def compress_external_weights(nus, regrets, rewards, lbd=0.0):
  """Computes distribution over `nus` while minimizing external regret.

  Args:
    nus: [T, P] array, T the number of different population distributions, P the
      number of different policies.
    regrets: [T, P] array, regrets[t, i] = payoff for switching from current
      policy to i at time t.
    rewards: [T, P] array, reward for playing policy P at time T.
    lbd: Sparsity argument.

  Returns:
    Computed distribution over `nus`.
  """

  def get_c(nus):
    return np.concatenate(
        (np.array([1.0]), -lbd * np.sum(rewards * nus, axis=1))
    )

  def get_max_constraints(nus, regrets, lbd):
    A = np.zeros((regrets.shape[1], 1 + nus.shape[0]))
    A[:, 0] = -1.0
    A[:, 1:] = np.transpose(
        regrets
        - np.sum(regrets * nus, axis=1).reshape(-1, 1)
        - lbd * np.abs(regrets)
    )
    return A, np.zeros(A.shape[0])

  def get_a_ub(nus, regrets, lbd):
    Amax, bmax = get_max_constraints(nus, regrets, lbd)
    Apos, bpos = get_proba_constraints_positivity(nus)
    return np.concatenate((Amax, Apos), axis=0), np.concatenate(
        (bmax, bpos), axis=0
    )

  c = get_c(nus)

  A_ub, b_ub = get_a_ub(nus, regrets, lbd)
  A_eq, b_eq = get_proba_constraint_sum_eq(nus)

  res = scipy.optimize.linprog(
      c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, options={'tol': 1e-10}
  )
  new_weights = res.x
  return new_weights[1:]
